#include "ChipCounterAlgorithm.h"
#include "AnalysisAlgorithmMap.h"
#include "SctData/ChipCounterTestResult.h"
#include "SctData/StandardDefects.h"
//#include "SctData/FitObject.h"
#include "SctData/RawScanResult.h"
#include <cfloat>
#include "TH2.h"
#include "Sct/SctNames.h"

// gcc296 needs a declaration of fabs
#include <cmath>
#include <set>

using namespace std;
using namespace boost;
using namespace SctData;
using namespace Sct;

namespace SctAnalysis {
    
    bool ChipCounterAlgorithm::inMap = AnalysisAlgorithmMap::instance().setAlgorithm("ChipCounterTest", auto_ptr<AnalysisAlgorithm>(new ChipCounterAlgorithm()));
    
    shared_ptr<AnalysisAlgorithm> ChipCounterAlgorithm::clone(const TestData& testData, const string& moduleName) const throw() {
	return shared_ptr<AnalysisAlgorithm>(new ChipCounterAlgorithm(testData, moduleName, *this));
    }

    void ChipCounterAlgorithm::loadData() {
	loadAllRaws();
    }
    
    bool ChipCounterAlgorithm::canAnalyze() const{
	return hasAllRaws();
    }
    
    shared_ptr<TestResult> ChipCounterAlgorithm::createTestResult() const {
	return shared_ptr<ChipCounterTestResult> (new ChipCounterTestResult());
    }
       
    void ChipCounterAlgorithm::analyze() {
      ChipCounterTestResult & result =  *dynamic_pointer_cast<ChipCounterTestResult> ( getTestResult() );
      
      const RawScanResult& raw = *getRaw(0);
      result.setScanVariable(raw.getHeader().getVariable());	

      const SctData::ScanPoints& points=raw.getPoints();
      const unsigned int numberOfScanPoints = points.getNPoints();
 
      const TH2& data0 = raw.getScanData(0);
      const TH2& data1 = raw.getScanData(1);

      const unsigned nx0 = data0.GetNbinsX();
      const unsigned ny0 = data0.GetNbinsY();
      const unsigned nx1 = data1.GetNbinsX();
      const unsigned ny1 = data1.GetNbinsY();
      if((!(ny0==numberOfScanPoints)) || (!(ny1==numberOfScanPoints))) { 
	throw IllegalStateError("This should not be!",__FILE__,__LINE__);
      };
      if((ny0<=0) || (ny1<=0) ||  (!(ny0==ny1))) {
	throw IllegalStateError("This should not be!",__FILE__,__LINE__);
      };

      // Find where the header bit starts 
      
      unsigned int i = 1;
      unsigned int nstart0 = 0;
      while ( i <= nx0 ) {
	if (data0.GetBinContent(i,1) == 0)  ++i;        
	if (data0.GetBinContent(i,1) == 1)  {
	  nstart0=i;
	  i=nx0+1;
	}        
      }

      i = 1;
      unsigned int nstart1 = 0;
      while ( i <= nx1 ) {
	if (data1.GetBinContent(i,1) == 0)  ++i;        
	if (data1.GetBinContent(i,1) == 1)  {
	  nstart1=i;
	  i=nx1+1;
	}        
      }
      
      // In this case, one of the links returned all 0's. How to deal with this?!?
      if ((nstart0 ==0) || (nstart1 ==0)) {
	throw IllegalStateError("This should not be!",__FILE__,__LINE__);
      }
      
      //      header: 111010 LLLL BBBBBBBB 101 = 21 bit header
      
      bool problemInBit[NBC+NL1];
      for (unsigned ibit=0; ibit<NBC+NL1; ++ibit){
	problemInBit[ibit]=false;
      }

      // Check the L1 and BC counters now
      for (int ibin=NHEADER; ibin<(NHEADER+NBC+NL1); ++ibin) { 
	unsigned int allzeros=1; //check that both counters have not failed to be all 0s
	
	for (unsigned int j=0; j< numberOfScanPoints; ++j) { 
	  if (!(data0.GetBinContent(nstart0+ibin, j+1) == data1.GetBinContent(nstart1+ibin, j+1))) {
	    // then the two counters are different. Report defect. 
	    problemInBit[ibin-NHEADER]=true;
	  }
	  if ((data0.GetBinContent(nstart0+ibin, j+1) == 1)) allzeros =0;
	}
	
	// also a defect if all zeros in that bin (need at least 15 triggers for this to be true!)
	if (allzeros == 1) problemInBit[ibin-NHEADER]=true;

      }
      
      bool setpassed=true;

      for (int ibit=0; ibit<NBC+NL1; ++ibit){
	if ( problemInBit[ibit] == true ) {
	  if (ibit < (NL1)) {
	    result.getDefects().addDefect(Defect(StandardDefects::L1_COUNTER, ModuleElement::Module()));
	    ostringstream oss;
	    oss << "Counter error in L1 bit (LSB=0) " << NL1-1-(ibit);  /// MSB is first
	    result.addComment(oss.str());
	    setpassed=false;
	  } else {
	    result.getDefects().addDefect(Defect(StandardDefects::BC_COUNTER, ModuleElement::Module()));
	    ostringstream oss;
	    oss << "Counter error in BC bit (LSB=0) " << NBC-1-(ibit-NL1);  /// MSB is first
	    result.addComment(oss.str());
	    setpassed=false;
	  }
	}
      }
      
      result.setPassed(setpassed);
      result.setProblem(!setpassed);
      
    }
} // end of namespace SctAnalysis
