#include "IOManagerDB.h"
#include "MySqlException.h"
#include "ZlibStringCompressor.h"
#include "Sct/Serializable.h"
#include "Sct/SctNames.h"
#include "Sct/XmlStyleStream.h"
#include "Sct/StdExceptionWrapper.h"
#include "Sct/Archive/IONameArchiveFile.h"

#include "Sct/UniqueID.h"
#include "SctData/ScanResult.h"
#include "SctData/TestResult.h"
#include <sstream>
#include <boost/date_time/posix_time/posix_time.hpp>

using std::string;
using std::ostringstream;
using boost::posix_time::to_simple_string;
using boost::shared_ptr;
using namespace Sct;

namespace SctArchiving {
  MysqlException::MysqlException(const string& msg, const string& file, int line){
    initialize("MYSQL_EXCEPTION", "ArchivingService::MysqlException", msg, 0, file, line);
  }

  MysqlException::MysqlException(const string& msg, const string& query, const string& file, int line) {
    std::ostringstream full_message;
    full_message << msg << "\nwhile executing query:\n" << query;
    initialize("MYSQL_EXCEPTION", "ArchivingService::MysqlException", full_message.str(), 0, file, line);
  }
  
  MysqlException::MysqlException(){}
  
  MysqlException::~MysqlException() throw(){}
  
  void IOManagerDB::checkForErrors(const string& msg, const string& file, int line){
    if(mysql_errno(m_mysql)){
      std::ostringstream full_message;
      full_message << msg << "\n" 
		   << mysql_error(m_mysql);
      std::cerr << "ERROR:" << full_message.str() << "\nat " << file << ":" << line << std::endl;
      throw MysqlException(full_message.str(), file, line);
    }
  }
  
  void IOManagerDB::checkForErrors(const string& msg, const string& query, const string& file, int line){
    if(mysql_errno(m_mysql)){
      std::ostringstream full_message;
      full_message << msg << "\n" 
		   << mysql_error(m_mysql);
      std::cerr << "ERROR:" << full_message.str()
		<< "\n executing: "
		<< query
		<< "\nat " << file << ":" << line << std::endl;
      throw MysqlException(full_message.str(), query, file, line);
    }
  }
  
  
  
  IOManagerDB::IOManagerDB() : m_locationName("nowhere") {
    std::cout <<"Creating IOManagerDB" << std::endl;
    m_compressor = shared_ptr<ZlibStringCompressor>(new ZlibStringCompressor());
    m_mysql = mysql_init((MYSQL*) 0);
    checkForErrors("Problem during initialisation of MYSQL structure.", __FILE__, __LINE__);
    
    // connect to the server
    std::cout <<"Connecting IOManagerDB" << std::endl;
    mysql_real_connect(m_mysql, "localhost", "daquser", "", NULL, 3306, NULL, 0);
    checkForErrors("Failed to connect to mysql server", __FILE__, __LINE__);
    SctNames::Mrs() << MRS_TEXT("Have connected to mysql server") << "IOMANAGER_DB_CONNECT" 
		    << MRS_INFORMATION << ENDM;

    // connect to the database
    std::string dbname="test_srd";
    std::cout <<"Connecting IOManagerDB to database name=`" << dbname << "'" << std::endl;
    mysql_select_db(m_mysql, dbname.c_str());
    std::string message("Failed to connect to database");
    message += dbname;
    checkForErrors(message, __FILE__, __LINE__);
    SctNames::Mrs() << MRS_TEXT("Have connected to mysql database") << "IOMANAGER_DB_CONNECT" 
		    << MRS_PARAM<const char*>("database_name",dbname.c_str())
		    << MRS_INFORMATION << ENDM;
  }
  
  IOManagerDB::~IOManagerDB(){
#warning "AJB This never seems to get called because the manager is a static beast. **Hopefully** this dosent matter!"
    std::string message("Closing connection to mysql server");
    std::cout << message << endl;
    SctNames::Mrs() << MRS_TEXT(message) << "IOMANAGER_DB_CONNECT" 
		    << MRS_INFORMATION << ENDM;
    mysql_close(m_mysql);
  }
  
  IOManagerDB* IOManagerDB::s_man=0;
  
  IOManagerDB& IOManagerDB::instance() {
    if (!s_man) s_man=new IOManagerDB();
    return *s_man;
  }
  
  std::string IOManagerDB::status() const {
    std::ostringstream oss;
    oss << "IOManagerDB";
    if (m_compressor.get()){
      oss << "String compressor buffer size : " 
	  << m_compressor->getBufferSize() << std::endl;
    } else {
      oss << "String compressor not initialized" << endl;
    }
    return oss.str();
  }

  const SctData::ResultHeader& IOManagerDB::getHeader(const Serializable& ob){
    const SctData::ResultHeader* header=0;
    if (ob.getClassName().find("TestResult")!=string::npos) {
      header = & dynamic_cast<const SctData::TestResult&>(ob).getHeader();
    } 
    if (ob.getClassName().find("ScanResult")!=string::npos) {
      header = & dynamic_cast<const SctData::ScanResult&>(ob).getHeader();
    }
    if (header) {
      return *header;
    } else {
      ostringstream oss;
      oss << "Dont know how to find a header for object with class name : " 
	  << ob.getClassName() << " which has UniqueID : " 
	  << (string) ob.getUniqueID();
      throw IoException(oss.str(), __FILE__, __LINE__);
    }
  }
  
  string IOManagerDB::getTable(const Serializable& ob){
    return "data";
  }

  void IOManagerDB::write(const Serializable& ob, const IOParams* par) const {
    try{
      boost::recursive_mutex::scoped_lock lock (getMutex());
      setReadMode(false);
      
    }catch (Sct::Throwable& e){
      throw;
    }catch(std::exception& e){
      StdExceptionWrapper sew(e);
      throw IoException(sew, __FILE__, __LINE__);
    }
  }

  boost::shared_ptr<Serializable> IOManagerDB::read(const string& name, const IOParams* par) const {
    try{
      boost::recursive_mutex::scoped_lock lock (getMutex());
      setReadMode(true);
      getReadVersionMap().clear();
      const string compressed;
#warning "AJB added null string above"
      std::istringstream iss(m_compressor->inflate(compressed));
      XmlStyleIStream in_ad(iss);
      return boost::dynamic_pointer_cast<Serializable>(readImpl(in_ad));      
    }catch (Sct::Throwable& e) {
      throw;
    }catch(std::exception& e){
      StdExceptionWrapper sew(e);
      throw IoException(sew, __FILE__, __LINE__);
    }
  }
  
bool IOManagerDB::testExists(const TestData& testdata){
  const std::string tableName="tblTests";
  std::ostringstream select;
  select << "SELECT tblTests.* FROM tblTestTypes " 
	 << "\nINNER JOIN ((tblLocations INNER JOIN tblRuns ON "
	 << "\ntblLocations.LocationID = tblRuns.LocationID) INNER JOIN tblTests ON "
	 << "tblRuns.RunID = tblTests.RunID) ON tblTestTypes.TestTypeID = tblTests.TestTypeID"
	 << "\nWHERE (((tblTestTypes.TestTypeName) = '" << testdata.testName << "' )\n" 
	 << "AND ((tblLocations.LocationName) = '" << getLocationName() << "')" 
	 << "\n AND ((tblRuns.RunNr)="  << testdata.runNumber
	 << ")\n AND ((tblTests.StartTime)='"
	 << convertToMysqlDateTimeFormat(testdata.startTime)
	 << "'))";
  
  mysql_query(m_mysql, select.str().c_str());
  checkForErrors("Could not check for test", select.str(), __FILE__, __LINE__);
  
  shared_ptr<MYSQL_RES> result (mysql_store_result(m_mysql));
  checkForErrors("Error storing result", select.str(), __FILE__, __LINE__);
  
  unsigned number_of_rows = mysql_num_rows(result.get());
  return (number_of_rows);
}


bool IOManagerDB::runExists(unsigned long runNumber){
  const std::string tableName="tblRuns";
  std::ostringstream select;
  select << "SELECT * FROM tblLocations INNER JOIN tblRuns "
	 <<" ON tblLocations.LocationID = tblRuns.LocationID "
	 <<" WHERE (((tblLocations.LocationName)='"
	 << getLocationName()
	 << "') AND ((tblRuns.RunNr)=" << runNumber << "))";
  mysql_query(m_mysql, select.str().c_str());
  
  checkForErrors("Could not check for run", select.str(), __FILE__, __LINE__);
  
  shared_ptr<MYSQL_RES> result (mysql_store_result(m_mysql));
  checkForErrors("Error storing result", __FILE__, __LINE__);
  
  unsigned number_of_rows = mysql_num_rows(result.get());
  return (number_of_rows);
}

void IOManagerDB::addRun(unsigned long runNumber){
  const std::string tableName="tblRuns";
  bool trash;
  std::ostringstream select;
  select << "INSERT INTO " << tableName << "\n("
	 << "LocationID, RunNr, Trash"
	 << ")"
	 << "\nSELECT tblLocations.LocationID" << ", "
	 << runNumber << " AS RunNr, 0 AS Trash "
	 << "\nFROM tblLocations"
	 << "\nWHERE ((tblLocations.LocationName) = '" 
	 << getLocationName()
	 << "')";
  mysql_query(m_mysql, select.str().c_str());
  checkForErrors("Cannot execute", select.str(), __FILE__, __LINE__);
  if (mysql_affected_rows(m_mysql)!=1){
    throw MysqlException("Did NOT add row", select.str(), __FILE__, __LINE__);
  }
}

std::string IOManagerDB::convertToMysqlDateTimeFormat(const std::string& original){
  std::string copy=original;
  unsigned t = copy.find('T');
  if (t==string::npos) {
    std::ostringstream msg;
    msg << "could not parse date/time since I couldn't find a `T' in " << copy;
    throw Sct::IoException(msg.str(), __FILE__, __LINE__);
  }
  return copy.erase(t,1);
}

void IOManagerDB::addTest(const TestData& testdata, bool alsoAddScans){
  const std::string tableName="tblTests";
  std::ostringstream select;

  select << "INSERT INTO " << tableName
	 << "\n(RunID, TestTypeID, StartTime, EndTime )"
	 << "\nSELECT tblRuns.RunID, tblTestTypes.TestTypeID, "
	 << "'" << convertToMysqlDateTimeFormat( testdata.startTime )
	 << "' AS StartTime, "
	 << "'" << convertToMysqlDateTimeFormat( testdata.endTime )
	 << "' AS EndTime ";
  select << "\nFROM " << "tblTestTypes" << ", " << "tblLocations"
	 << " INNER JOIN " << "tblRuns" << " ON "
	 << "tblLocations.LocationID = tblRuns.LocationID";
  select << "\nWHERE ((" 
	 << "(tblTestTypes.TestTypeName) = '" <<  testdata.testName << "')"
	 << "\nAND ((tblRuns.RunNr) = " << testdata.runNumber << ")"
	 << "\nAND ((tblLocations.LocationName) = '" << getLocationName() << "'))";

  mysql_query(m_mysql, select.str().c_str());
  checkForErrors("Cannot execute", select.str(), __FILE__, __LINE__);
  if (mysql_affected_rows(m_mysql)!=1){
    throw MysqlException("Did NOT add row", select.str(), __FILE__, __LINE__);
  }
  if (alsoAddScans){
    for (unsigned iscan=testdata.startScanNumber; 
	 iscan < (testdata.startScanNumber + testdata.nScans);
	 ++iscan){
      addScan(testdata, iscan);
    }
  }
}

bool IOManagerDB::scanExists(const TestData& testdata, unsigned scanNumber){
#warning "NOT IMPLIMENTED"
  return false;
}

void IOManagerDB::addScan(const TestData& testdata, unsigned scanNumber){
  const std::string tableName="tblScans";
  std::ostringstream select;
  
  select << "INSERT INTO " << tableName
	 << "\n(TestID, ScanNr)"
	 << "\nSELECT tblTests.TestID, " << scanNumber
	 << " AS ScanNr ";
  select << "\nFROM " << "tblTestTypes INNER JOIN (" << "tblLocations"
	 << " INNER JOIN (" << "tblRuns INNER JOIN "
         << "tblTests ON tblRuns.RunID = tblTests.RunID)"
         << " ON tblLocations.LocationID = tblRuns.LocationID)"
         << " ON tblTestTypes.TestTypeID = tblTests.TestTYpeID";
  select << "\nWHERE ((" 
	 << "(tblTestTypes.TestTypeName) = '" <<  testdata.testName << "')"
	 << "\nAND ((tblRuns.RunNr) = " << testdata.runNumber << ")"
	 << "\nAND ((tblLocations.LocationName) = '" << getLocationName() << "'))"
         << "\nAND ((tblTests.StartTime) = '" << convertToMysqlDateTimeFormat(testdata.startTime) << "')";
  
  mysql_query(m_mysql, select.str().c_str());
  checkForErrors("Cannot execute", select.str(), __FILE__, __LINE__);
  if (mysql_affected_rows(m_mysql)!=1){
    throw MysqlException("Did NOT add row", select.str(), __FILE__, __LINE__);
  }
}

void IOManagerDB::setLocationName(const std::string& name){
  m_locationName=name;
}

std::string IOManagerDB::getLocationName(){
  return m_locationName;
}

}
