
#ifndef SCT_LOOKUP_H
#define SCT_LOOKUP_H



#include <string>
#include "ipc/partition.h"
#include "is/isinfo.h"
#include "SctNames.h"
#include "MultiMessageDebugStream.h"

namespace Sct {

       /**
	* There is a lot of rubbish in this file, but the main idea
	* is as follows:
	*
	* Sergei's IPCPartition::lookup<> method is very lazy.
	* It returns IPC references from a cache under most
	* circumstances, and only does THEY JOB IT OUGT TO DO
	* SOMETIMES (i.e. if the user were to request it to!)
	* when it realises that its cache is invalid, which
	* very annoyingly only discovers when you try to USE
	* one of the bum references it gives you.
	*
	* What we really want is a lookup method that gives you
	* the OPTION of using or not using the cache.  The purpose
	* of this particular file is to provide an alternative
	* Sct::Lookup method that wraps Sergei's one, but which does
	* some tests on the returned IPC reference first.
	*
	* In principle, the idea is that it returns a reference THAT 
	* IS VALID at the time of the call (though of course it might
	* become invalid only seconds later if you are unlucky).
	*
	* If a valid reference to the object you are asking for can't
	* be found, then this method returns TT::_nil().
	*
	* ------------ Stop reading if you are not a developer ----
	* Experimentally it is noted that if the asked for object had died
	* and been reborn, then the following algorithm
	* executes the while loop 3 times while getting rid of Sergei's
	* strange policy:
	*   ntries=0:
	*   	NIL=true, NXT=true.
	*   ntries=1;
	*   	NIL=false, NXT=true.
	*   ntires=2:
	*   	NIL=false, NXT=false.
	*   (then the algorithm returns)
	* Note that the only exceptions that are (believed to be caught)
	* are those commented "/// KNOWN TO BE CAUGHT" below. 
	*/	

	
  template < class TT >
  typename TT::_ptr_type Lookup(const std::string & name, const IPCPartition & p){
    const bool PRINT_DEBUG = true;
    
    typename TT::_ptr_type ob = TT::_nil();
    bool valid=false;
    int ntries=0;
    const int maxtries=10; // must be >= 2
    
    while (valid==false && ntries<maxtries){

      if (PRINT_DEBUG) {
        MultiMessageDebugStream m(true,false,false);
        m << "Starting attempt " << ntries;
        m.flush();
      };
      
      const bool NIL=CORBA::is_nil(ob);
      bool NXT=true;
      try {
        if (PRINT_DEBUG) {
          MultiMessageDebugStream m(true,false,false);
          m << "about to test existence";
          m.flush();
        };
        NXT=ob->_non_existent();
      } catch (const ipc::CacheExpired & e) {
	/// KNOWN TO BE CAUGHT
	NXT=true; // Since it appears not to exist yet !   :(
      };
      
      if (PRINT_DEBUG) {
        MultiMessageDebugStream m(true,false,false);
        m << "thingkinsdf";
        m.flush();
      };
      
      if (NIL || NXT) {
	if (PRINT_DEBUG) {
	  MultiMessageDebugStream m(true,false,false);
   	  m << "At attempt "<<ntries<<" it was found that NIL="<<NIL<<" and NXT="<<NXT<<" so trying to lookup again!";
	  m.flush();
        };	  
	try {
          ob = (p.template lookup<TT>( name ));
	  if (PRINT_DEBUG) {
	    MultiMessageDebugStream m(true,false,false);
	    m << "No exception here, fortunately";
	    m.flush();
	  };
	} catch (const ipc::CacheExpired & e) {
	  /// NOT KNOWN TO BE CAUGHT
 	  ob = TT::_nil(); // let's try again
	  if (PRINT_DEBUG) {
	    MultiMessageDebugStream m(true,false,false);
	    m << "Good job we caught that!";
	    m.flush();
	  };
	};
      } else {
	if (PRINT_DEBUG) {
	  MultiMessageDebugStream m(true,false,false);
   	  m << "At attempt "<<ntries<<" we think we have a good answer.";
	  m.flush();
	};	  
        valid=true;
      }
      ++ntries;
    }
    if (!valid) {
      if (PRINT_DEBUG) {
        Sct::MultiMessageDebugStream m(true,true,false);
        m << "Our lookup wrapper failed to look up ["<<name<<"] after "<<maxtries<<" attempts.";
        m.flush();
      };
      return TT::_nil();
    } else {
      if (PRINT_DEBUG) {
	MultiMessageDebugStream m(true,false,false);
   	m << "Lookup returning something successfully";
	m.flush();	
      };
      return ob;
    }
  };
  
};
#endif
