#include <CommonWithDsp/sctStructure.h>
#include <ABCD/ABCDscans.h>
#include <ABCD/ABCDchip.h>
#include "extraScans.h"
#include "SctApiFwd.h"

#include <iostream>

namespace SctApi{
/** compares ABCDModule structs and their sub-types */
class ConfigUtility{
 public:
  /** 
      check equality of ABCDBasic ignoring any unused bits 
      @return bitfield with bits of ConfigType which are different
  */
  static unsigned diff(const ABCDBasic& first, const ABCDBasic& second);
  /** 
      check equality of ABCDChip ignoring any unused bits 
      @return bitfield with bits of ConfigType which are different
  */
  static unsigned diff(const ABCDConfig& first, const ABCDConfig& second);
  /** 
      check equality of ABCDChip ignoring any unused bits 
      @return bitfield with bits of ConfigType which are different
   */
  static unsigned diff(const ABCDCaldata& first, const ABCDCaldata& second);
  /** 
      check equality of ABCDChip ignoring any unused bits 
      @return bitfield with bits of ConfigType which are different
  */
  static unsigned diff(const ABCDChip& first, const ABCDChip& second);
  /** 
      check equality of ABCDModule ignoring any unused bits 
      @return bitfield with bits of ConfigType which are different
  */
  static unsigned diff(const ABCDModule& first, const ABCDModule& second);
  /**
     Clone a module
  */
  static std::auto_ptr<ABCDModule> clone(const ABCDModule& first);
  /** 
      Clone the basic part of a chip
  */
  static std::auto_ptr<ABCDBasic> clone(const ABCDBasic& first);

  /**
     set an ABCD module variable:
   */
  static void modifyVar(ABCDModule* module, UINT32 typ, FLOAT32 val);
  /**
     set an ABCD module chip variable:
   */
  static void modifyVar(ABCDModule* module, unsigned chip, UINT32 typ, FLOAT32 val);

  /**
     return true if the type is a BOC register rather than a ABCD register
   */
  static bool isBOCRegister(UINT32 typ);

  enum ConfigVarType {
    BOC_CHANNEL_CONFIG_VAR, BOC_GLOBAL_CONFIG_VAR,
    MODULE_GLOBAL_CONFIG_VAR, MODULE_CHIP_CONFIG_VAR, MODULE_CHANNEL_CONFIG_VAR,
    UNKNOWN_CONFIG_VAR
  };

  /**
     return true if the type is a BOC register rather than a ABCD register
   */
  static ConfigVarType variableType(UINT32 typ);
 private:
  static ConfigUtility& ConfigUtility::instance();
  ConfigUtility();
  UINT16 config_compare_mask;
  static void setToken(ABCDModule* module, unsigned token, int link0, int link1);
  static INT32 ConfigUtility::setChipRole(ABCDModule *module, UINT32 theChip, ABCD_ROLES role);
};

inline std::auto_ptr<ABCDModule> ConfigUtility::clone(const ABCDModule& first){
  std::auto_ptr<ABCDModule> m (new ABCDModule);
  const size_t size=sizeof(ABCDModule);
  memcpy(m.get(), &first, size);
  return m;
}

inline std::auto_ptr<ABCDBasic> ConfigUtility::clone(const ABCDBasic& first){
  std::auto_ptr<ABCDBasic> a (new ABCDBasic);
  const size_t size=sizeof(ABCDBasic);
  memcpy(a.get(), &first, size);
  return a;
}

inline
ConfigUtility::ConfigUtility(){
  config_compare_mask=0xffff;
  reinterpret_cast<ABCDConfig*>(&config_compare_mask)->padding=0;
}

inline
ConfigUtility& ConfigUtility::instance(){
  static ConfigUtility config;
  return config;
}

// compare the first 14 bits only, since the last 2 are padding 
inline unsigned ConfigUtility::diff(const ABCDConfig& first, const ABCDConfig& second){
  unsigned diff_bits=SCTAPI_CONFIG_NONE;
  const UINT16* p=reinterpret_cast<const UINT16*>(&first);
  const UINT16* q=reinterpret_cast<const UINT16*>(&second);
  const UINT16& cfg_mask=instance().config_compare_mask;
  
  UINT16 cfgdiff = *p^*q;
  cfgdiff &= cfg_mask;
  if (cfgdiff) {
    diff_bits |= 1<<SCTAPI_CONFIG_CFG;
  }
  return diff_bits;
}

inline unsigned ConfigUtility::diff(const ABCDBasic& first, const ABCDBasic& second){
  unsigned diff_bits=diff(first.config, second.config);

  if  (first.vthr!=second.vthr || first.vcal!=second.vcal ||
       first.delay!=second.delay || first.preamp!=second.preamp ||
       first.shaper!=second.shaper) diff_bits|=1<<SCTAPI_CONFIG_BASIC;
  
  for (unsigned i=0; i<4; ++i){
    if (first.mask[i]!=second.mask[i]) diff_bits|=1<<SCTAPI_CONFIG_MASK;
  }
  return diff_bits;
}

inline unsigned ConfigUtility::diff(const ABCDCaldata& first, const ABCDCaldata& second){
  unsigned diff_bits=SCTAPI_CONFIG_NONE;

  if (first.rc_function!=second.rc_function || 
      first.c_factor!=second.c_factor) diff_bits |= 1<<SCTAPI_CONFIG_OTHER;

  for (unsigned i=0; i<3; ++i){
    if (first.rc_params[i]!=second.rc_params[i]) diff_bits |= 1<< SCTAPI_CONFIG_OTHER;
  }

  return diff_bits;
}

inline unsigned ConfigUtility::diff(const ABCDChip& first, const ABCDChip& second){
  unsigned diff_bits=SCTAPI_CONFIG_NONE;
  
  diff_bits |= diff(first.basic, second.basic);
  diff_bits |= diff(first.caldata, second.caldata);

  for (unsigned i=0; i<N_SCT_CHANS; ++i){
    if (first.trim[i]!=second.trim[i]) diff_bits |= 1<<SCTAPI_CONFIG_TRIM;
  }

  if (first.active!=second.active ||
      first.address!=second.address ||
      first.target!=second.target ) diff_bits |= 1<<SCTAPI_CONFIG_OTHER;  
  return diff_bits;
}

inline unsigned ConfigUtility::diff(const ABCDModule& first, const ABCDModule& second){
  unsigned diff_bits=SCTAPI_CONFIG_NONE;
  
  for (unsigned ichip=0; ichip<N_SCT_CHIPS; ++ichip){
    diff_bits |= diff(first.chip[ichip], second.chip[ichip]);
  }
  
  if (first.present!=second.present || first.active!=second.active ||
      first.select!=second.select || first.groupId!=second.groupId ||
      first.pTTC!=second.pTTC || first.rTTC!=second.rTTC ||
      first.rx[1]!=second.rx[1] || first.rx[2]!=second.rx[2]) {
    diff_bits |= 1<<SCTAPI_CONFIG_OTHER;
  }
  return diff_bits;
}


inline bool ConfigUtility::isBOCRegister(UINT32 typ){
  switch(typ) {
  case ST_RX_DELAY:
  case ST_RX_DELAY0:
  case ST_RX_DELAY1:
  case ST_RX_THRESHOLD:
  case ST_RX_THRESHOLD0:
  case ST_RX_THRESHOLD1:
  case ST_TX_CURRENT:
  case ST_TX_MARKSPACE:
  case ST_TX_DELAY:
  case ST_TX_COARSE:
  case ST_TX_FINE:
  case SCT_SCAN_BOC_BPM_PHASE:
  case SCT_SCAN_BOC_BREG_PHASE:
  case SCT_SCAN_BOC_V0_PHASE:
  case SCT_SCAN_BOC_V1_PHASE:
  case SCT_SCAN_BOC_V_PHASES:
  case SCT_SCAN_BOC_VRN_FINE:
  case SCT_SCAN_TX_CHANNELS:
  case SCT_SCAN_RAW_TX_CHANNELS:
    return true;
  default:
    return false;
  }
}

}
