
#ifndef SCT_SCTAPI_SCAN_H
#define SCT_SCTAPI_SCAN_H

/**
   Scan class

   Describes a scan request

   There are two module sets defined by the module group 
   number in the configuration.
*/ 

//fwd dec
namespace SctApi {
  class Scan;
};

// includes
#include <vector>
#include <boost/shared_ptr.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <CommonWithDsp/processor.h>
#include "ModuleList.h"

// other fwd decs
namespace SctApi {
  class Trigger;
};

//defn
namespace SctApi {
  class Scan {
  public:
    virtual ~Scan() {};

    /// Typedef for list of scan points
    typedef std::vector<FLOAT32> ScanPoints;

    /// Typedef for list of scan points
    typedef std::vector<UINT32> TrigPoints;

    /// Current options for a scan request
    /**
       FULL: if 1 decode three time bins into consecutive histograms

       BITS32: 0 -> 16 bits  1 -> 32 bits

       LOOPCALLINE: Do a loop over the calibration line for each bin

       DISTSLAVE: 0 -> All to slave 0  1 -> Group distribution  2 -> Router distribution

       DEBUG: If scan doesn't complete drop into debug mode

       TIM: Triggers sent by TIM

       NTH: Histogram every nth trigger (only in router distribution mode)

       NTH_REM: Remainder for nth trigger

       ENABLE_DATA_MODE: Normally true.  Says that when the module is configured, the module will go into data-returning mode.  The alternative (false) puts the module in return-configuration-register-id mode.

    */
    typedef enum ScanOptions {FULL, BITS32, LOOPCALLINE, DISTSLAVE, DEBUG, TIM, NTH, NTH_REM, ENABLE_DATA_MODE} ScanOptionsEnum;

    virtual void configure(unsigned short type, float start, float stop, float step) = 0;
    virtual void configure2(unsigned short type, float start, float stop, float step) = 0;

    virtual const boost::shared_ptr<Trigger> getTrigger1() const = 0;
    virtual const boost::shared_ptr<Trigger> getTrigger2() const = 0;

    virtual boost::shared_ptr<Trigger> getTrigger1() = 0;
    virtual boost::shared_ptr<Trigger> getTrigger2() = 0;

    virtual void setTrigger1(boost::shared_ptr<Trigger> trigger) = 0;
    virtual void setTrigger2(boost::shared_ptr<Trigger> trigger) = 0;

    virtual void setScanVariable1(unsigned short var) = 0;
    virtual void setScanVariable2(unsigned short var) = 0;

    virtual unsigned short getScanVariable1() const = 0;
    virtual unsigned short getScanVariable2() const = 0;

    virtual void setNTrigs(unsigned long nTrigs) = 0;
    virtual unsigned long getNTrigs() const = 0;

    virtual const TrigPoints getVariableTrigs() const = 0;
    virtual void setVariableTrigs(const TrigPoints &scans) = 0;
    virtual void setVariableTrigRange(unsigned short start, unsigned short end, unsigned long value) = 0;

    virtual const ScanPoints getScanPoints1() const = 0;
    virtual const ScanPoints getScanPoints2() const = 0;
    virtual void setScanPoints1(const ScanPoints &scans) = 0;
    virtual void setScanPoints2(const ScanPoints &scans) = 0;

    virtual int getOption(enum ScanOptions) const = 0;
    virtual void setOption(enum ScanOptions, int option) = 0;

    /// Print the scan
    virtual void print() const = 0;

    virtual void setScanNumber(unsigned int scan) = 0;
    virtual unsigned int getScanNumber() const = 0;
    virtual void setRunNumber(unsigned int scan) = 0;
    virtual unsigned int getRunNumber() const = 0;

    virtual ModuleList getModuleList(unsigned int group) const = 0;
    virtual void setModuleList(unsigned int group, const ModuleList & newList) = 0;
    virtual unsigned int getNGroups() const = 0;
    virtual void setNGroups(unsigned int val) = 0;
  
    virtual boost::posix_time::ptime getStartTime() const = 0;
    virtual void setStartTime(boost::posix_time::ptime t) = 0;
  
    virtual boost::posix_time::ptime getEndTime() const = 0;
    virtual void setEndTime(boost::posix_time::ptime t) = 0;
  };
};

#endif
