
#ifndef SCT_SCTAPI_SCTAPIDDC_H
#define SCT_SCTAPI_SCTAPIDDC_H

/************ Direct DDC access (no translations) ****************/

// fwddec
namespace SctApi {
  class SctApiDDC;
};

// includes
#include <string>
#include <map>
#include <vector>
#include "CommonWithDsp/processor.h"
#include "boost/shared_ptr.hpp"

// other fwddecs
class IPCPartition;
class SCTDCSCommand;
namespace SctConfiguration {
  class Configuration;
};
namespace SctApi{
  class Log;
};

// defn
namespace SctApi {
  class SctApiDDC {
  private:
    /// I think I have to do this after ILU is running (otherwise SCTDCSCommand can't do the stuff it needs to initialise)
    void initDDC();
    /// The configuration database
    boost::shared_ptr<SctConfiguration::Configuration> m_config;
    boost::shared_ptr<Log> m_log;
    IPCPartition * m_ddcPartition;
    std::string * m_ddcCtrls[2];
    SCTDCSCommand * m_ddcCmd;
    /// safe private ddc accessor:
    SCTDCSCommand * ddcCmd();
    /// safe private config accessor:
    SctConfiguration::Configuration * config();
    std::ostream & log();
  public:
    SctApiDDC(boost::shared_ptr<SctConfiguration::Configuration> conf,
              boost::shared_ptr<Log> log = boost::shared_ptr<Log>());
    /// Change state on one channel
    void ddcChangeState (int crate, int channel, short state, int timeout = 2);
    /// Changeto LV on/off, HV on/off on one channel, or group, or everything
    void ddcChangeStates (int crate, const char* name, short state, int timeout = 2);
    /// Change state on all power supply channels
    void ddcChangeStateAll (int crate, short state, int timeout = 2);
    /// Load power supply crate configuration for particular state
    void ddcLoadConfiguration (int crate, short state, int timeout = 2);
    /// Send hard reset to power supply channel
    void ddcHardReset (int crate, int channel, int timeout = 2);
    /// Send hard reset to all channels in power supply crate
    void ddcHardResetAll (int crate, int timeout = 2);

    /// Get a parameter that should be interpreted as an integer
    int ddcGetChannelParameterInt (int crate, int channel, std::string name, int timeout = 2);
    /// Get a parameter that should be interpreted as a float
    float ddcGetChannelParameterFloat (int crate, int channel, std::string name, int timeout = 2);
    /// Get a parameter that should be interpreted as a string (most likely an error)
    std::string ddcGetChannelParameterString (int crate, int channel, std::string name, int timeout = 2);
    /// Get a card parameter that should be interpreted as an integer
    int ddcGetCardParameterInt (int crate, int card, std::string name, int timeout = 2);
    /// Get a card parameter that should be interpreted as a float
    float ddcGetCardParameterFloat (int crate, int card, std::string name, int timeout = 2);
    /// Get a card parameter that should be interpreted as a string (probably an error)
    std::string ddcGetCardParameterString (int crate, int card, std::string name, int timeout = 2);
    /// Get a parameter of a power supply crate that should be interpreted as an integer
    int ddcGetCrateParameterInt (int crate, std::string name, int timeout = 2);
    /// Get a parameter of a power supply crate that should be interpreted as a float
    float ddcGetCrateParameterFloat (int crate, std::string name, int timeout = 2);
    /// Get a parameter of a power supply crate that should be interpreted as a string
    std::string ddcGetCrateParameterString (int crate, std::string name, int timeout = 2);

    /// Get some parameters of a power supply crate (interpreted as floats)
    /** 
        @param crate The crate id
        @param names A list of strings representing parameters
        @param dpeValues The return value, map parameter name strings to values
    */
    void ddcGetCrateParameters(int crate, std::vector<std::string> names, 
                               std::map<std::string, float>& dpeValues, int timeout);

    /// Set a parameter on a power supply channel
    void ddcSetChannelParameter (int crate, int channel, std::string name, float value, int timeout = 2);
    /// Set a parameter on a power supply card
    void ddcSetCardParameter (int crate, int card, std::string name, float value, int timeout = 2);
    /// Set a parameter on a power supply crate
    void ddcSetCrateParameter (int crate, std::string name, float value, int timeout = 2);

    /************ NOW THE STUFF WHICH USE EXTRA FUNCTIONALITY *************/
    
    /*** DCS functionality */
    
    /**
       Request hard reset on all modules

       DDC:  Primary purpose
       IPC:  Not during scan
       Conf: Some lookup?!
    */
    void requestHardResetAll();

    /**
       Request hard reset of module mid

       DDC:  Primary purpose
       IPC:  Not during scan
       Conf: Some lookup?!
    */
    void requestHardReset(UINT32 mid);

    /**
       Request initialisation of an IV scan

       DDC:  Primary purpose
       IPC:  Not during scan, result will be asynchronous
       Conf: Some lookup?!
    */
    void requestIVCurve(UINT32 mid, FLOAT32 start, FLOAT32 stop, FLOAT32 step, UINT16 delay, FLOAT32 currentLim);

    /**
       Request initialisation of an IV scan on all modules

       DDC:  Primary purpose
       IPC:  Not during scan, result will be asynchronous
       Conf: Some lookup?!
    */
    void requestIVCurveAll(FLOAT32 start, FLOAT32 stop, FLOAT32 step, UINT16 delay, FLOAT32 currentLim);

    /**
       Set the select value for all known power supply channels in a crate

       This is a debug method for DCS. Primary route to this functionality is 
       via modifyABCDVar and sendModule???

       DDC:  Primary purpose
       IPC:  Not during scan
       Conf: Some lookup?!
    */
    void setSelectAllInCrate(int crate, bool value, int timeout = 5);

    /**
       Set the select value for module mid

       This is a debug method for DCS. Primary route to this functionality is 
       via modifyABCDVar and sendModule!

       DDC:  Primary purpose
       IPC:  Not during scan
       Conf: Some lookup?!
    */
    // #warning "Should this just change DCS?"
    void setSelect(UINT32 mid, bool value); 

    /**
       Change run state of all power modules

       One of Off, Standby, On

       DDC:  Primary purpose
       IPC:  Not during scan
       Conf: No communication
    */
    void changeRunState(int state);

  };
};

#endif
