/**
   SctApiMConf.cxx

   This file has methods which do things to module configurations on the ROD.
   Mainly, where primitives exist to make these modifications.

   Set and Get configuration between the ROD and SctApi memory.
   Send configuration from ROD to the modules.

   Changes to module configurations in SctApi memory (cache) are in SctApiModify.cxx
 */

#include "primListWrapper.h"
#include "PrimBuilder.h"
#include "SctApi.h"
#include "SctApiRodInfo.h"
#include "crate.h"
#include "utility.h"
#include "SctApiDebug.h"

#include "CommonWithDsp/primParams.h"
#include "CommonWithDsp/registerIndices.h"

#include "Sct/FibreNumberConverters.h"
#include "ConfigurationUtility.h"
#include "sctConf/configuration.h"
#include "SctApiException.h"

using namespace std;
using namespace SctPixelRod;

namespace SctApi {

void SctApi::getABCDModules(BankType bank_type) {
  {
    boost::mutex::scoped_lock lock(log().mutex());
    log() << "getABCDModule (all). Bank: " << bank_type << endl;
  }
  for(map<UINT32, ABCDModule> ::const_iterator iter = moduleMap.begin();
      iter != moduleMap.end();
      iter ++) {
    getABCDModule(iter->first, bank_type);
  }
}


void SctApi::getABCDModule(UINT32 mid, BankType bank_type) {
  boost::shared_ptr<SctApiConfigCache::ModuleBanks> the_banks=module_cache.getFromMid(mid);
  if (!the_banks.get()) {
    if(mrs) {
      *mrs << "ROD_BANK_CONFIG_CACHE" << MRS_ERROR << MRS_QUALIF("SCTAPI") 
           << MRS_PARAM<int>("mid", mid)
           << MRS_TEXT("Cant return cache for non-existant MID") << ENDM;
    }
    return;
  }
  boost::shared_ptr<ABCDModule> cached_config = the_banks->get(bank_type);
  if (!cached_config.get()){
    if(mrs) {
      *mrs << "ROD_BANK_CONFIG_CACHE" << MRS_ERROR << MRS_QUALIF("SCTAPI") 
           << MRS_PARAM<int>("mid", mid)
           << MRS_PARAM<int>("bank", bank_type)
           << MRS_TEXT("Cant return cache for non-existant BANK") << ENDM;
    }
    return;
  }

  ABCDModule *previous = lookupConfig(mid);
  if(previous) {
    *previous = *cached_config;
  } else {
    moduleMap.insert(make_pair(mid, *cached_config));
  }

  // option to test cache of this bank against the real ROD bank:
  if(checkDebugOption(DEBUG_API_CONFIG_CACHE_CHECK)){
    previous = lookupConfig(mid);
    if (!previous) throw SctApiException("SctApiMConf: null pointer error which should never happen");
    std::auto_ptr<ABCDModule> previous_clone = ConfigUtility::clone(*previous);
    really_getABCDModule(mid, bank_type);

    ABCDModule *current = lookupConfig(mid);

    // get bit-field of difference types
    unsigned differences=ConfigUtility::diff(*previous_clone, *current);
    if (differences){
      if (mrs){
	*mrs << "ROD_BANK_CONFIG_CACHE" << MRS_WARNING << MRS_QUALIF("SCTAPI") 
	     << MRS_PARAM<int>("module", mid)
	     << MRS_TEXT("DEBUG error - configuration in CACHE and ROD differ!")             << MRS_PARAM<int>("diff bitfield", differences)
             << ENDM;
      }
      boost::mutex::scoped_lock lock(log().mutex());
      log() << "ERROR on module configuration cache:"
	    << " module" <<  mid << " bitfield of differences:" << std::hex << differences 
	    << std::dec << std::endl;
      log() << "Printing CACHE and then ROD to configuration service output"<< std::endl;
      config->printModuleConfig(*previous_clone);
      config->printModuleConfig(*current);
    }
  }
}

void printBanks(std::list<BankType> banks, std::ostream& os){
   for (std::list<BankType>::const_iterator i=banks.begin(); i!=banks.end(); ++i){
      if (i!=banks.begin()) os << ",";
      os << *i;
   }
   if (banks.begin()==banks.end()) os << "NONE!";
}

void SctApi::copyABCDModules(BankType source, std::list<BankType> to){
#warning "AJB - this should be changed to use multiple copyABCDModule when it changes"
  std::cout << "AJB AJB copyABCDModules from " << source << " to ";
  printBanks(to, std::cout); 
  std::cout << endl;
  getABCDModules(source);
  setABCDModules(to);
}

void SctApi::copyABCDModule(UINT32 mid, BankType source, std::list<BankType> to_banks){
#warning "AJB - this should be changed to use the new ROD primitive ASAP!"
  std::cout << "Using non-primitive version of copyABCDModule" << std::endl;
  getABCDModule(mid, source);
  setABCDModule(mid, to_banks);  
  
  std::cout << "AJB AJB copyABCDModule mid=" << mid << " from " << source << " to ";
  printBanks(to_banks, std::cout);
  std::cout << endl;
  
  return;
  /*
    AJB - code below is awaiting a copy primitive ...
    until then the two methods above (get/set) simulate this
  */
#warning "AJB awaiting a copy primitive"
  
  boost::shared_ptr<SctApiConfigCache::ModuleBanks> the_banks=module_cache.getFromMid(mid);
  if (!the_banks.get()) {
    if(mrs) {
      *mrs << "ROD_BANK_CONFIG_CACHE" << MRS_ERROR << MRS_QUALIF("SCTAPI") 
	   << MRS_PARAM<int>("mid", mid)
	   << MRS_TEXT("Cant return cache for non-existant MID") << ENDM;
    }else{
      std::cout << "Cant return cache for non-existant MID " << mid << std::endl;
    }
    return;
  }
  boost::shared_ptr<ABCDModule> from_config = the_banks->get(source);
  if (!from_config.get()){
    if(mrs) {
      *mrs << "ROD_BANK_CONFIG_CACHE" << MRS_ERROR << MRS_QUALIF("SCTAPI") 
	   << MRS_PARAM<int>("module", mid)
	   << MRS_PARAM<int>("bank", source)
	   << MRS_TEXT("Cant return cache for non-existant BANK") << ENDM;
    }else{
      std::cout << "Cant return cache for non-existant source BANK " << source << std::endl;
    }
    return;
  }
  std::list<BankType> modify_bank_list;
  for (std::list<BankType>::const_iterator bank_iterator=to_banks.begin(); 
       bank_iterator!=to_banks.end(); ++bank_iterator){

    BankType to_bank=*bank_iterator;

    boost::shared_ptr<ABCDModule> to_config = the_banks->get(to_bank);
    if (!to_config.get()){
      to_config=boost::shared_ptr<ABCDModule>(ConfigUtility::clone(*from_config).release());
      the_banks->set(to_bank, to_config);
      modify_bank_list.push_back(to_bank);
    }else{
      unsigned differences=ConfigUtility::diff(*from_config, *to_config);
      if (to_bank==SCTAPI_BANK_SCAN) differences = SCTAPI_CONFIG_ALL;
      if (differences) {
	*to_config=*from_config;
	modify_bank_list.push_back(to_bank);
      }
    }
  }

  /** toy prim list */

  unsigned int partition, crate, rod, channel;
  Utility::getpcrc(mid, partition, crate, rod, channel);

  for (list<BankType>::const_iterator bank_iterator=modify_bank_list.begin();
       bank_iterator!=modify_bank_list.end(); ++ bank_iterator){
    int realBank = Utility::translateBank(*bank_iterator);
    // add translation of bank to multi-bank thingy
  }
  boost::shared_ptr<PrimListWrapper> primList(new PrimListWrapper(1));

  sendPrimList(rod, primList);

  int responseCode = awaitResponse(rod, 10);
  
  if(responseCode!=0) {
    cout << "Copy module unsuccessful!\n";
  }
}

void SctApi::setABCDModule(UINT32 mid, std::list<BankType> bank_list) {
  std::cout << "AJB AJB setABCDModule mid=" << mid << " banks=";
  printBanks(bank_list, std::cout);
  std::cout << std::endl;
  bool config_is_new=false;
    
  using boost::shared_ptr;
  ABCDModule* lookup=lookupConfig(mid);
  if (!lookup) {
    std::cerr << "setABCDModule : Could not find module config for mid = " << mid << std::endl;
    return;
  }

#warning "TODO: fix so send configuration using different ROD if necessary"
  /* Look up using offRODTTC to find ROD and virtual channel number
     this should already be in the module configuration cache */

  shared_ptr<SctApiConfigCache::ModuleBanks> the_banks=module_cache.getFromMid(mid);

  // get cache of banks for this mid, or create it (in which case the config must be sent)
  if (!the_banks.get()) {
    config_is_new=true;
    the_banks=shared_ptr<SctApiConfigCache::ModuleBanks>(new SctApiConfigCache::ModuleBanks);
    module_cache.set(mid, the_banks);
  }
  

  std::list<BankType> modified_bank_list;
  for (std::list<BankType>::const_iterator bank_iterator=bank_list.begin(); 
       bank_iterator!=bank_list.end(); ++bank_iterator){
    BankType bank_type=*bank_iterator;
    std::cout << "AJB AJB on bank " << bank_type << std::endl;
    shared_ptr<ABCDModule> current_cache=the_banks->get(bank_type);
    
    // get ABCDModule for this bank, or create it (in which case the config must be sent)
    if (!current_cache.get()) {
      current_cache = shared_ptr<ABCDModule>(ConfigUtility::clone(*lookupConfig(mid)).release());
      the_banks->set(bank_type, current_cache);
      modified_bank_list.push_back(bank_type);
    }else{
      /*
	Get bit-field of all differences between local config and ROD cache.
	1/3/2005 AJB has set the conditional below because at the end of the scan 
	the SCAN bank is in an odd state, hence we always should copy the scan bank,
	if that has been explicitly requested
      */
      unsigned differences=ConfigUtility::diff(*current_cache, *lookup);
      if (bank_type==SCTAPI_BANK_SCAN) differences = SCTAPI_CONFIG_ALL;
      if (differences){
	*current_cache = *lookup;
	modified_bank_list.push_back(bank_type);
      }else{
	std::cout << "AJB AJB No need to set ABCDModule type " << bank_type
		  << " for mid = " << mid
		  << " and bank_type  = " << bank_type
		  << " as configuration hasnt changed, " << endl;
	if(checkDebugOption(DEBUG_API_CONFIG_CACHE_SEND)){
	  std::cout << "Debug option set so sending module configuration ANYWAY!" << std::endl;
	  modified_bank_list.push_back(bank_type);
	}
      }
    }
  }
  really_setABCDModule(mid, modified_bank_list);
}

void SctApi::really_getABCDModule(UINT32 mid, BankType bank_type) {
  {
    boost::mutex::scoped_lock lock(log().mutex());
    log() << "getABCDModule " << mid << " Bank: " << bank_type << endl;
  }

  // Find module channel in ROD
  unsigned int partition, crate, rod, channel;
  Utility::getpcrc(mid, partition, crate, rod, channel);

  // Which bank id to use
  int realBank = Utility::translateBank(bank_type);

  boost::shared_ptr<PrimListWrapper> primList(new PrimListWrapper(1));
  PrimBuilder::instance().readModuleData(primList, realBank, channel);

  sendPrimList(rod, primList);

  int responseCode = awaitResponse(rod, 5);

  unsigned long length;
  unsigned long *outBody = 0;
  if(responseCode == 0) {
    outBody = getResponse(rod, length);
  }

  if(outBody) {
//      unsigned long outLength = outBody[0];
//      unsigned long outIndex = outBody[1];
    unsigned long outNumPrims = outBody[2];
//      unsigned long outPrimVersion = outBody[3];

    if(outNumPrims == 1) {
      unsigned long primLength = outBody[4];
//        unsigned long primIndex = outBody[5];
      unsigned long primID = outBody[6];
//        unsigned long primVersion = outBody[7];

      if(primID == RW_MODULE_DATA) {
        if(primLength != (sizeof(ABCDModule)/sizeof(UINT32) + 4)) { // Prim length includes header
          cout << "Output primitive was " << primLength << " not " << sizeof(ABCDModule)/sizeof(UINT32) << endl;
        }

        // Data is at outBody[8]
        ABCDModule result;

        // Copy data
        for(unsigned int i=0; i<sizeof(ABCDModule)/sizeof(UINT32); i++) {
          ((unsigned long *)&result)[i] = outBody[8 + i];
        }

        ABCDModule *previous = lookupConfig(mid);

        if(previous) {
          cout << "Replacing previous module config\n";

          *previous = result;
        } else {
          cout << "New module config inserted\n";
          moduleMap.insert(make_pair(mid, result));
        }
      } else {
        cout << "Returned primitive was not RW_MODULE_DATA\n";
      }
    } else {
      cout << "Too many primitives in output list\n";
    }

    delete [] outBody;
  } else {
    cout << "No response to primitive list!\n";
  }
}

boost::shared_ptr<ABCDModule> SctApi::getABCDModuleRaw(unsigned int rod, 
                                                       UINT32 slot, BankType bank_type) {
  {
    boost::mutex::scoped_lock lock(log().mutex());
    log() << "getABCDModuleRaw Slot: " << slot << " Bank: " << bank_type << endl;
  }

  boost::shared_ptr<ABCDModule> result;

  // Which bank id to use
  int realBank = Utility::translateBank(bank_type);

  boost::shared_ptr<PrimListWrapper> primList(new PrimListWrapper(1));
  PrimBuilder::instance().readModuleData(primList, realBank, slot);
  sendPrimList(rod, primList);

  int responseCode = awaitResponse(rod, 5);

  unsigned long length;
  unsigned long *outBody = 0;
  if(responseCode == 0) {
    outBody = getResponse(rod, length);
  }

  if(outBody) {
    unsigned long outNumPrims = outBody[2];

    if(outNumPrims == 1) {
      unsigned long primLength = outBody[4];
      unsigned long primID = outBody[6];

      if(primID == RW_MODULE_DATA) {
        unsigned int expectedSize = (sizeof(ABCDModule)/sizeof(UINT32) + 4);
        if(primLength != expectedSize) { // Prim length includes header
          cout << "Output primitive was " << primLength << " not " << expectedSize << endl;
        }

        result.reset(new ABCDModule);

        // Data is at outBody[8]
        // Copy data
        for(unsigned int i=0; i<sizeof(ABCDModule)/sizeof(UINT32); i++) {
          ((unsigned long *)result.get())[i] = outBody[8 + i];
        }
      } else {
        cout << "Returned primitive was not RW_MODULE_DATA\n";
      }
    } else {
      cout << "Too many primitives in output list\n";
    }

    delete [] outBody;
  } else {
    cout << "No response to primitive list!\n";
  }

  return result;
}

/* Store config in ROD */
void SctApi::setABCDModules(std::list<BankType> banks) {
  std::cout << "setABCDModules (all) banks=";
  printBanks(banks, std::cout);
  std::cout << std::endl;
  {
    boost::mutex::scoped_lock lock(log().mutex());
    log() << "setABCDModule called for all modules. Banks : ";
    for (std::list<BankType>::const_iterator it=banks.begin(); it!=banks.end(); ++it){
      log() << *it <<",";
    }
    log() << endl;
  }
  for(map<UINT32, ABCDModule>::const_iterator iter = moduleMap.begin();
      iter != moduleMap.end();
      iter ++) {
    cout << "Set Config " << iter->first << endl;
    setABCDModule(iter->first, banks);
  }

  // Off-ROD modules are already in the moduleMap so get sent automatically
  return;
}

/* Store config in ROD */
void SctApi::really_setABCDModule(UINT32 mid, std::list<BankType> bank_list) {
  std::cout << "really_setABCDModule mid=" << mid << " banks=";
  printBanks(bank_list, std::cout);
  std::cout << std::endl;
  if(mid==0xffffffff){
    cout << "setABCDModule with 0xffffffff\n";
    throw SctApiException("mid out of range");
  }


  std::string moduleName;
  ABCDModule *moduleConfig;
  
  moduleConfig = lookupConfig(mid);
  
  if(!moduleConfig) {
    cout << "*** NO module configuration for mid = " << mid << " in set module" << endl;
    return;
  }
  
  cout << "\nModule config\n";
  cout << " primary channel " << (int)moduleConfig->pTTC 
       << " red channel " << (int)moduleConfig->rTTC 
       << " group " << (int)moduleConfig->groupId
       << " select " << (int)moduleConfig->select << endl;
  cout << " target (chip 0) " << (int)moduleConfig->chip[0].target << endl;

  for (std::list<BankType>::const_iterator bank_iterator=bank_list.begin();
       bank_iterator!=bank_list.end(); ++bank_iterator){
    BankType bank_type=*bank_iterator; 
    {
      boost::mutex::scoped_lock lock(log().mutex());
      log() << "AJB AJB setABCDModule " << mid << " Bank: " << bank_type << endl;
    }
    std::cout << "AJB AJB setABCDModule " << mid << " Bank: " << bank_type << endl; 
    unsigned int partition, crate, rod, channel;
    Utility::getpcrc(mid, partition, crate, rod, channel);
    
    int realBank = Utility::translateBank(bank_type);
    
    boost::shared_ptr<PrimListWrapper> primList(new PrimListWrapper(1));
    PrimBuilder::instance().writeModuleData(primList, moduleConfig, realBank, channel);
    
    RodLabel rl(partition, crate, rod);
    
    for(int l=0; l<2; l++) {
      const unsigned int zeroTo123Pos = moduleConfig->rx[l];
      // We check the values before using it, doing this here makes the next bit less confusing
      const unsigned int zeroTo95Pos = Sct::FibreNumberConverters::from124FormatTo96Format(zeroTo123Pos);
      if (zeroTo123Pos==DATA_LINK_OFF) {
	// Don't try to get non-existant value!
	//  but otherwise this is valid ... no problem
      } else if(zeroTo123Pos <= 123 && zeroTo95Pos < 96) {
	// This is already cached in cacheModuleConfig
	const unsigned int val = getRodInfo(rl).linkErrorMasks[zeroTo95Pos];
	PrimBuilder::instance().writeRegister(primList, ERROR_MASK(zeroTo95Pos/48, zeroTo95Pos%48), 0, 13, val);
        std::cout << "ERROR_MASK: writing value " << val  << " to " << ERROR_MASK(zeroTo95Pos/48, zeroTo95Pos%48) << endl; 
      } else {
	std::ostringstream os;
	os << "Bad fibre number for error mask values! Line " << __LINE__ << " of " << __FILE__ << " when 123pos = " << zeroTo123Pos <<" and 95pos = " << zeroTo95Pos;
	if (mrs) {
	  *mrs << "SctApi" << MRS_ERROR << MRS_TEXT(os.str()) << ENDM;
	} else {
	  std::cout << "SctApi: ERROR " << os.str() << std::endl;
	  return;
	}
      }
    }
    
#if USE_DUAL_PORTS
    // Set up module masks so config can be sent using both serial ports
    setupModuleMask(SP_BOTH, 0xf, primList);
#else
    // This will be a null op if it doesn't exist
#ifdef CMD_BUFFER_0
#warning "Use of old header files detected"
    setupModuleMask(CMD_BUFFER_0, 0xf, primList);
#else
    setupModuleMask(SP0, 0xf, primList);
#endif
#endif
    sendPrimList(rod, primList);

    int responseCode = awaitResponse(rod, 10);

    if(responseCode!=0) {
      cout << "Set module unsuccessful!\n";
    }
  }
}


void SctApi::sendAllABCDModules(BankType bank_type,
                                ConfigType config_type,
                                const bool enableDataMode /* enableDataMode should have default value true to ensure compatibility with old code*/) {

  std::cout << "AJB AJB sendAllABCDModules bank=" << bank_type << " config_type="
            << config_type << " enableDataMode=" << enableDataMode << std::endl;
  {
    boost::mutex::scoped_lock lock(log().mutex());
    log() << "sendABCDModule called for all modules. Bank: " << bank_type << endl;
  }

  /*
    This method is far too long...

    1) Change to ROD mode that allows configuration to be sent to the modules
    2) Check the BOC isn't interlocked
    3) Build and send a primitive list
     a) Disable trigger decoder
     b) Disable formatters
     c) Send configuration
     d) Reenable trigger decoder
    4) Change "back" to CALIBRATION_MODE

    option 2 (for early versions of the DSP code, probably not every stage necessary)
    1) Disable formatters
    2) Send configuration for each module in turn using sendABCDModule
    3) Setup module masks
    4) Reset some FIFOs...

*/

  cout << "Send config to all modules\n";

  PrimBuilder &builder = PrimBuilder::instance();

  // Set ROD mode to something appropriate (may be in TIM mode!)
  boost::shared_ptr<PrimListWrapper> rodModeList1(new PrimListWrapper(1));
  // Should CALIBRATION_SLINK_OVERRIDE_MODE be added to this (it works without!)
  builder.rodMode(rodModeList1, CALIBRATION_MODE, 0, 1, 1, 1, 1);

  synchSendPrimListAllCrates(rodModeList1);


  // If new primitive use it instead of doing a loop
#if (R_SEND_CONFIG == 104) || (R_SEND_CONFIG == 105)
  {
    boost::mutex::scoped_lock lock(log().mutex());
    log() << "\tUsing one SEND_CONFIG primitive\n";
  }

  {
    for(list<RodLabel>::const_iterator rl = rodList.begin();
        rl!=rodList.end();
        rl++){

      if(!getCrate()->checkBOCLasersOn(rl->rod)) {
        cout << "Trying to send module configuration using BOC that has its lasers cut out\n";
        if(mrs) {
          *mrs << "BOC_INTERLOCKED" << MRS_ERROR << MRS_QUALIF("SCTAPI") 
               << MRS_PARAM<int>("crate", rl->crate) 
               << MRS_PARAM<int>("rod", rl->rod)
               << MRS_TEXT("Can't send all module config (BOC interlocked)") << ENDM;
        }

        throw SctApiException("Send modules aborted due to interlock status");
      }
    }
  }

  boost::shared_ptr<PrimListWrapper> sendConfigList(new PrimListWrapper(1));

  // Write "Disable trigger decoder" to primList
#ifdef RRIF_CMND_1
  builder.writeRegister(sendConfigList, RRIF_CMND_1, 18, 1, 0);
#else
#error "Unsupported no registers"
#endif   // RRIF_CMND_1

  // Disable all formatters
#ifdef FMT_LINK_EN
  for(int i=0; i<8; i++) {
    builder.writeRegister(sendConfigList, FMT_LINK_EN(i), 0, 32, 0);
  }
#else
#warning "Probably don't need to turn off formatters!?"
#endif // FMT_LINK_EN

  int realBank = Utility::translateBank(bank_type);
  int realConfig = Utility::translateConfigType(config_type);
  //@todo SP_BOTH doesn't work with current DSP code
#if USE_DUAL_PORTS
  // Send module configuration from both serial ports (ie two modules at a time)
  builder.sendConfig(sendConfigList, SP_BOTH, 0, ALL_MODULES, ALL_MODULES, ALL_CHIPS, 1, 1,
                     realBank, MODULE_GROUP_ALL, realConfig, 0, enableDataMode);
#else
  builder.sendConfig(sendConfigList, SP0, 0, ALL_MODULES, ALL_MODULES, ALL_CHIPS, 1, 1,
                     realBank, MODULE_GROUP_ALL, realConfig, 0, enableDataMode);
#endif

  // Write "Re-enable trigger decoder" to primList
#ifdef RRIF_CMND_1
  builder.writeRegister(sendConfigList, RRIF_CMND_1, 18, 1, 1);
#else
#error "Unsupported no registers"
#endif  // RRIF_CMND_1

  // setupModuleMask to reenable appropriate formatters??

  sendPrimListAll(sendConfigList);
  
  {
    int responseCode = awaitResponseAll(10);
    
    if(responseCode != 0) {
      cout << "Send modules unsuccessful\n";
    }
  }

  boost::shared_ptr<PrimListWrapper> rodModeList(new PrimListWrapper(1));
  // Should CALIBRATION_SLINK_OVERRIDE_MODE be added to this (it works without!)
  builder.rodMode(rodModeList, CALIBRATION_MODE, 0, 1, 1, 1, 1);

  sendPrimListAll(rodModeList);

  {
    int responseCode = awaitResponseAll(10);
    
    if(responseCode != 0) {
      cout << "Send modules rod mode unsuccessful\n";
    }
  }
#else   // Use slow config (loop over modules)
  {
    boost::mutex::scoped_lock lock(log().mutex());
    log() << "\tUsing multiple SEND_CONFIG primitives\n";
  }

#warning "Slow module sending"

  // Disable all formatters to avoid bus error!
#ifdef FMT_LINK_EN
  boost::shared_ptr<PrimListWrapper> clearFormatList(new PrimListWrapper(1));

  for(int i=0; i<8; i++) {
    builder.writeRegister(clearFormatList, FMT_LINK_EN(i), 0, 32, 0);
  }

  sendPrimListAll(clearFormatList);
  
  {
    int responseCode = awaitResponseAll(10);
    
    if(responseCode != 0) {
      cout << "Clear formatters unsuccessful\n";
    }
  }
#else
#warning "Probably don't need to turn off formatters! (old version)"
#endif   // FMT_LINK_EN

  for(map<UINT32, ABCDModule>::const_iterator iter = moduleMap.begin();
      iter != moduleMap.end();
      iter ++) {
    cout << "SendConfig " << iter->first << endl;
    sendABCDModule(iter->first, bank_type, config_type, enableDataMode);
  }
#endif   // End of slow config sending

  // Restore link masks etc
  boost::shared_ptr<PrimListWrapper> setupMasks(new PrimListWrapper(1));

#if USE_DUAL_PORTS
  // Masks also have to be both serial ports
  setupModuleMask(SP_BOTH, 0xf, setupMasks);
#else
#ifdef CMD_BUFFER_0
  setupModuleMask(CMD_BUFFER_0, 0xf, setupMasks);
#else
  setupModuleMask(SP0, 0xf, setupMasks);
#endif   // CMD_BUFFER_0
#endif

  sendPrimListAll(setupMasks);
  
  {
    int responseCode = awaitResponseAll(10);
    
    if(responseCode != 0) {
      cout << "Setup masks unsuccessful\n";
    }
  }
  // Reset everything so its not confused by the clk/2
  boost::shared_ptr<PrimListWrapper> resetList(new PrimListWrapper(1));

  cout << "Resetting front end interface\n";
#ifdef RRIF_CMND_0
  builder.writeRegister(resetList, RRIF_CMND_0, 0, 8, 0xff);  // Reset everything
  builder.writeRegister(resetList, RRIF_CMND_0, 0, 8, 0x78);  // Leave debug FIFO in reset
#else
#error "Unsupported no registers"
#endif   // RRIF_CMND_0

  sendPrimListAll(resetList);

  int responseCode = awaitResponseAll(10);

  if(responseCode != 0) {
      cout << "ROD front end reset unsuccessful\n";
  }

  return;
}

  void SctApi::sendABCDModule(UINT32 mid,
			      BankType bank_type, 
			      ConfigType config_type, 
			      const bool enableDataMode) {
    /* enableDataMode should have default value true to ensure compatibility with old code*/
  if(mid==0xffffffff){
    cout << "sendABCDModule with 0xffffffff\n";
    throw SctApiException("mid out of range");
  }
  
  int realConfig=Utility::translateConfigType(config_type);

  cout << "Send config to module " << mid << endl;

  {
    boost::mutex::scoped_lock lock(log().mutex());
    log() << "sendABCDModule: Mid: " << mid << " Bank: " << bank_type << " Type: " << config_type << endl;
  }

  unsigned int partition, crate, rod, channel;
  Utility::getpcrc(mid, partition, crate, rod, channel);

  if(!getCrate()) {
    cout << "Send module for non-existent crate " << partition << " " << crate << endl;
    return;
  }

  int realBank = Utility::translateBank(bank_type);

  if(realBank > SPARE_MODULE_CONFIG) {
    cout << "sendABCDModule with illegal bank (translated) " << realBank << endl;
    throw SctApiException("Illegal Bank");
  }

  if(!getCrate()->checkBOCLasersOn(rod)) {
    cout << "Trying to send module configuration using BOC that has its lasers cut out\n";
    if(mrs) {
      *mrs << "BOC_INTERLOCKED" << MRS_ERROR << MRS_QUALIF("SCTAPI") 
           << MRS_PARAM<int>("module", mid) << MRS_PARAM<int>("crate", crate) 
           << MRS_PARAM<int>("rod", rod)
           << MRS_TEXT("Can't send module config (BOC interlocked)") << ENDM;
    }

    return;
  }

  boost::shared_ptr<PrimListWrapper> primList(new PrimListWrapper(1));
  PrimBuilder &builder = PrimBuilder::instance();

  // Write "Disable trigger decoder" to primList
#ifdef RRIF_CMND_1
  builder.writeRegister(primList, RRIF_CMND_1, 18, 1, 0);
#else
#error "Unsupported no registers"
#endif

  // Seeing as we're only sending one module, use only one serial port
  builder.sendConfig(primList, SP0, 0, channel, NO_MODULE, ALL_CHIPS, 1, 1,
                     realBank, MODULE_GROUP_ALL, realConfig, 0, enableDataMode);

  // Write "Re-enable trigger decoder" to primList
#ifdef RRIF_CMND_1
  builder.writeRegister(primList, RRIF_CMND_1, 18, 1, 1);
#else
#error "Unsupported no registers"
#endif

  sendPrimList(rod, primList);

  int responseCode = awaitResponse(rod, 10);

  if(responseCode != 0) {
    cout << "Send module unsuccessful\n";
  }
}

#warning "AJB fix to use primitive for multiple banks"
void SctApi::modifyABCDVarROD(UINT32 mid, UINT32 chip, UINT32 type, FLOAT32 value, std::list<BankType> banks) {
  {
    boost::mutex::scoped_lock lock(log().mutex());
    log() << "modifyABCDVarROD (module " << mid << ": chip " << chip  << ") " << type << " " << value << " banks";
    printBanks(banks, log());
    log() << std::endl;
  }

  cout << "Change variable macro\n";

  unsigned int partition, crate, rod, channel;
  Utility::getpcrc(mid, partition, crate, rod, channel);

  if(!isRODPresent(rod)) {
    cout << "Trying to change module configuration on a non existant ROD\n";
    if(mrs) {
      *mrs << "ROD_UNCONFIGURED" << MRS_ERROR << MRS_QUALIF("SCTAPI") 
           << MRS_PARAM<int>("module", mid) << MRS_PARAM<int>("crate", crate) 
           << MRS_PARAM<int>("rod", rod)
           << MRS_TEXT("Can't modify module config (ROD non-existant)") << ENDM;
    }

    return;
  }

  for (std::list<BankType>::const_iterator bank_iterator=banks.begin(); bank_iterator!=banks.end(); ++bank_iterator){
    BankType bank_type=*bank_iterator;

    std::cout << "AJB AJB modifyABCDVarROD doing bank " << bank_type << std::endl;
    
    boost::shared_ptr<PrimListWrapper> primList(new PrimListWrapper(1));
    int realBank = Utility::translateBank(bank_type);
    
    PrimBuilder::instance().writeModuleVariable(primList, realBank, MODULE_GROUP_ALL, channel, chip, type, value);
    
    sendPrimList(rod, primList);
    awaitResponse(rod, 5);
    
    // also update the cache!
    // update cache...
    boost::shared_ptr<SctApiConfigCache::ModuleBanks> the_banks=module_cache.getFromMid(mid);
    if (!the_banks.get()) {
      if(mrs) {
	*mrs << "ROD_BANK_CONFIG_CACHE" << MRS_ERROR << MRS_QUALIF("SCTAPI") 
	     << MRS_PARAM<int>("module", mid) << MRS_PARAM<int>("crate", crate) 
	     << MRS_PARAM<int>("rod", rod)
	     << MRS_TEXT("Cant update cache for non-existant MID") << ENDM;
      }else{
	std::cout << "Cant update cache for non-existant MID "<< mid;
      }
      return;
    }
    boost::shared_ptr<ABCDModule> cached_config = the_banks->get(bank_type);
    if (!cached_config.get()){
      if(mrs) {
	*mrs << "ROD_BANK_CONFIG_CACHE" << MRS_ERROR << MRS_QUALIF("SCTAPI") 
	     << MRS_PARAM<int>("module", mid) << MRS_PARAM<int>("crate", crate) 
	     << MRS_PARAM<int>("rod", rod)
	     << MRS_PARAM<int>("bank", bank_type)
	     << MRS_TEXT("Cant update cache for non-existant BANK") << ENDM;
      }else{
	std::cout << "Cant update cache for non-existant BANK " << bank_type << std::endl;
      }
      return;
    }
    ConfigUtility::modifyVar(cached_config.get(), chip, type, value);
  }
}
  
#warning "AJB fix to use primitives for multiple banks"
void SctApi::modifyABCDVarROD(UINT32 mid, UINT32 type, FLOAT32 value, std::list<BankType> banks) {
  cout << "Change variable macro\n";
  
  unsigned int partition, crate, rod, channel;
  Utility::getpcrc(mid, partition, crate, rod, channel);
  
  if(!isRODPresent(rod)) {
    cout << "Trying to change module configuration on a non existant ROD\n";
    if(mrs) {
      *mrs << "ROD_UNCONFIGURED" << MRS_ERROR << MRS_QUALIF("SCTAPI") 
           << MRS_PARAM<int>("module", mid) << MRS_PARAM<int>("crate", crate) 
           << MRS_PARAM<int>("rod", rod)
           << MRS_TEXT("Can't modify module config (ROD non-existant)") << ENDM;
    }
    
    return;
  }
  
  for (std::list<BankType>::const_iterator bank_iterator=banks.begin(); bank_iterator!=banks.end(); ++bank_iterator){
    BankType bank_type=*bank_iterator;  
    {
      boost::mutex::scoped_lock lock(log().mutex());
      log() << "modifyABCDVarROD (module " << mid << ") " << type << " " << value << " bank_type " << bank_type << "\n";
    }    
    
    boost::shared_ptr<PrimListWrapper> primList(new PrimListWrapper(1));
    
    int realBank = Utility::translateBank(bank_type);
    
    PrimBuilder::instance().writeModuleVariable(primList, realBank, MODULE_GROUP_ALL, channel, ALL_CHIPS, type, value);
    
    sendPrimList(rod, primList);
    awaitResponse(rod, 5);
    
    // update cache...
    boost::shared_ptr<SctApiConfigCache::ModuleBanks> the_banks=module_cache.getFromMid(mid);
    if (!the_banks.get()) {
      if(mrs) {
	*mrs << "ROD_BANK_CONFIG_CACHE" << MRS_ERROR << MRS_QUALIF("SCTAPI") 
	     << MRS_PARAM<int>("module", mid) << MRS_PARAM<int>("crate", crate) 
	     << MRS_PARAM<int>("rod", rod)
	     << MRS_TEXT("Cant update cache for non-existant MID") << ENDM;
      }
    }
    boost::shared_ptr<ABCDModule> cached_config = the_banks->get(bank_type);
    if (!cached_config.get()){
      if(mrs) {
	*mrs << "ROD_BANK_CONFIG_CACHE" << MRS_ERROR << MRS_QUALIF("SCTAPI") 
	     << MRS_PARAM<int>("module", mid) << MRS_PARAM<int>("crate", crate) 
	     << MRS_PARAM<int>("rod", rod)
	     << MRS_TEXT("Cant update cache for non-existant BANK") << ENDM;
      }
    }
    ConfigUtility::modifyVar(cached_config.get(), type, value);
  }
}


#warning "AJB fix to use primitives for multiple banks"
void SctApi::modifyABCDVarROD(UINT32 type, FLOAT32 value, std::list<BankType> banks) {
  cout << "Change variable macro\n";


  for (std::list<BankType>::const_iterator bank_iterator=banks.begin(); bank_iterator!=banks.end(); ++bank_iterator){
    BankType bank_type=*bank_iterator;
    std::cout << "modifyABCDVarROD bank " << bank_type << std::endl;
    {
      boost::mutex::scoped_lock lock(log().mutex());
      log() << "modifyABCDVarROD (all modules, all chips) " << type << " " << value << " Bank: " << bank_type << "\n";
    }

    boost::shared_ptr<PrimListWrapper> primList(new PrimListWrapper(1));
    
    int realBank = Utility::translateBank(bank_type);
    
    PrimBuilder::instance().writeModuleVariable(primList, realBank, MODULE_GROUP_ALL, ALL_MODULES, ALL_CHIPS, type, value);
    
    for(list<RodLabel>::const_iterator rl = rodList.begin();
	rl!=rodList.end();
	rl++){
      
      unsigned int rod = rl->rod;
      
      if(!isRODPresent(rod)) {
	cout << "Trying to change module configuration on a non existant ROD\n";
	if(mrs) {
	  *mrs << "ROD_UNCONFIGURED" << MRS_ERROR << MRS_QUALIF("SCTAPI") 
	       << MRS_PARAM<int>("rod", rod)
	       << MRS_TEXT("Can't modify module config (ROD non-existant)") << ENDM;
	}
	
	return;
      }
      sendPrimList(rod, primList);
      awaitResponse(rod, 2);
    }
  }
}

} // Close namespace
