#include <unistd.h>
#include <stdlib.h>
#include <iostream>

#include <mrs/message.h>

#include <algorithm>
#include <sstream>
#include <memory>

#include "Sct_SctApi/SctApi.hh"
#include "SctApiServer.h"
#include "sctConfIPC/configipc.h"
#include "SctApi/autoConfig.h"

#include "Sct/MultiMessageDebugStream.h"
#include "Sct/SctNames.h"
#include "Sct/OmniMarshalling.h"
#include "Sct/DestroyingDeleter.h"
#include "SctApi/Idiosyncrasy.h"
#include "sctConfIPC/marshalling.h"
#include "marshalling.h"

#include "SctApi/SctApi.h"

#include "TriggerWrapper.h"
#include    "ScanWrapper.h"

#include "Sct/StringStreamer.h"

using namespace Sct;
using namespace std;

#include "ScanServer.h"
#include "TriggerServer.h"
#include "ScanStatusManager.h"
#include "SctApi/SctApiDDC.h"
#include "boost/bind.hpp"

/*
  CORBA::ULong is long unsigned int
 */

SctApiServer *SctApiServer::singleton = 0;

SctApiServer::SctApiServer(const CrateDescriptors & apis, 
                           const char* interfaceName) : 
  IPCNamedObject<POA_Sct_SctApi::SctApiIPC>(Sct::SctNames::getPartition(),
                                            interfaceName),
  m_assm(boost::shared_ptr<Sct::ScanStatusManager>
         (new Sct::ScanStatusManager())),
  scans(0), 
  defaultScan(0),
  defaultTrigger(0),
  mrsStream(Sct::SctNames::getPartition()), 
  // NOT NEEDED ddcSingleTimeout(3), ddcMultipleTimeout(15),

  m_crateStore()
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  for (CrateDescriptors::const_iterator it = apis.begin();
       it != apis.end();
       ++it) {
    const std::string crateControllerName = SctApi::Idiosyncrasy(*it).ipcNameOfSctApiCrateServer();

    IPCPartition & p = Sct::SctNames::getPartition();
    {
      Sct::MultiMessageDebugStream m(true,false,true);
      m << "SctApiServer attempting to contact crate controller for " << *it << " using ipc name [" << crateControllerName <<"]";
    }
    Sct_SctApi::CrateIPC_var crateController = p.lookup<Sct_SctApi::CrateIPC>(crateControllerName);

    if (CORBA::is_nil(crateController)) {
      std::ostringstream s;
      s << "Could not find crate controller called ["<< crateControllerName << "] while contructing the top SctApiServer.  Aborting at line " << __LINE__ << " in file " << __FILE__ << ".  Are you sure that you remembered to start the \"ipc_server\" and the \"ipc_server -p SCT\" and the crate controller itself ?";
      try {
        Sct::MultiMessageDebugStream m(true,true,true);
        m << s.str();
        m.flush();
      } catch (...) {
        // do nothing;
      };
      throw Sct_SctApi::SctApiException(s.str().c_str());
    };
    this->m_crateStore.registerNewCrateController(crateController);
  };

  defaultScan = new ScanServer;
  //REMOVED api.setMrsStream(&mrsStream);
  //REMOVED api.setIsDictionary();

  mrsStream << "SCTAPI_CONSTRUCT" << MRS_INFORMATION << MRS_QUALIF("SctApiServer") << MRS_TEXT("Constructor called") << ENDM;

  try {
    publish(); // publish should be late in the constructor so that we don't have to remember to withdraw() in the event that something following publish() were to throw!
  } catch (CORBA::Exception & e) {
    try {
      Sct::MultiMessageDebugStream m(true,true,true);
      m << "CORBA::Exception [" << e._name() << "] when publishing SctApiServer in " << __FILE__<<" at line " << __LINE__;
      m.flush();
    } catch (...) {
      // fall through
    };
    throw;
  };
  singleton = this;
}

SctApiServer::~SctApiServer() 
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  mrsStream << "SCTAPI_DESTRUCT" << MRS_INFORMATION << MRS_QUALIF("SctApiServer") << MRS_TEXT("Destructor called") << ENDM;

  singleton = 0;

  withdraw();
  //deactivate();

  std::cout << "Forcefully deleting " << scans.size() << " scans in TopServer\n";
  scans.clear();
  std::cout << "Forcefully deleted " << scans.size() << " scans in TopServer\n";
  /*
    // This old method is (hopefuly) no longer needed now that elements of "scans" have custom delters!
    for(Scans::iterator i = scans.begin();
      i!= scans.end(); 
      i++) {
    delete *i;
  }
  */

#warning LESTER SUSPECTS THAT THE AUTHORS OF THIS ROUTINE SHOULD ALSO BE FREEING THINGS LIKE THE DEFAULT SCAN AND THE DEFAULT TRIGGER!
}

void SctApiServer::initialiseAll (CORBA::ULong runNumber)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  try {
    mrsStream << "SCTAPI_INIT" << MRS_INFORMATION << MRS_QUALIF("SctApiServer") << MRS_TEXT(Sct::StringStreamer<< "initialiseAll called with run number "<< runNumber) << ENDM;

    for (Workers::iterator wit = workers().begin();
	 wit != workers().end();
	 ++wit) {
      (*wit)->initialiseAll(runNumber);
    };

  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by initialiseAll: " << s.detail << endl;
    throw;
  }
}

void SctApiServer::shutdownAll ()
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  try {
    mrsStream << "SCTAPI_SHUTDOWN" << MRS_INFORMATION << MRS_QUALIF("SctApiServer") << MRS_TEXT("shutdownAll called") << ENDM;

    for (Workers::iterator wit = workers().begin();
	 wit != workers().end();
	 ++wit) {
      (*wit)->shutdownAll();
    }
    
  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by shutdownAll: " << s.detail << endl;
    throw;
  }
}

void SctApiServer::status ()
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  for (Workers::iterator wit = workers().begin();
       wit != workers().end();
       ++wit) {
    (*wit)->status();
  }
}

void SctApiServer::awaitResponse (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  try {
    worker()->awaitResponse(partition, crate, rod);
  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by awaitResponse: " << s.detail << endl;
    throw;
  }
}

Sct_SctApi::DataBlock * SctApiServer::getResponse (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  try {
    return worker()->getResponse(partition, crate, rod);
  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by getResponse: " << s.detail << endl;
    throw;
  }
}

void SctApiServer::configureBOC (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  try {
    worker()->configureBOC(partition, crate, rod);
  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by configureBOC: " << s.detail << endl;
    throw;
  }
}

Sct_SctApi::BOCMonitorSequence * SctApiServer::getBOCMonitorArray (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  try {
    return worker()->getBOCMonitorArray(partition, crate, rod);
  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by getBOCMonitorArray: " << s.detail << endl;
    throw;
  }
}

void SctApiServer::flashLED (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::UShort slaveNumber, CORBA::ULong period, CORBA::ULong flashes)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  try {
    worker()->flashLED(partition, crate, rod, slaveNumber, period, flashes);
  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by flashLED: " << s.detail << endl;
    throw;
  }
}

void SctApiServer::echo (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, const Sct_SctApi::DataBlock & data)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  try {
    worker()->echo(partition, crate, rod, data);
  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by echo: " << s.detail << endl;
    throw;
  }
}

void SctApiServer::createDebugPrimList ()
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  try {
    worker()->createDebugPrimList();
  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by createDebugPrimList: " << s.detail << endl;
    throw;
  }
}

void SctApiServer::addDebugPrimList (CORBA::Long index, CORBA::Long id, CORBA::Long version, const Sct_SctApi::DataBlock & body)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  try {
    worker()->addDebugPrimList(index, id, version, body);
  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by addDebugPrimList: " << s.detail << endl;
    throw;
  }
}

void SctApiServer::sendDebugPrimList (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  try {
    worker()->sendDebugPrimList(partition, crate, rod);
  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by sendDebugPrimList: " << s.detail << endl;
    throw;
  }
}

void SctApiServer::sendDebugSlavePrimList (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::ULong slave, CORBA::Boolean await, CORBA::Boolean response)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  try {
    worker()->sendDebugSlavePrimList(partition, crate, rod, slave, await, response);
  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by sendDebugSlavePrimList: " << s.detail << endl;
    throw;
  }
}

void SctApiServer::setRunNumber (CORBA::ULong newRun)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  for (Workers::iterator wit = workers().begin();
       wit != workers().end();
       ++wit) {
    (*wit)->setRunNumber(newRun);
  };
}

void SctApiServer::setScanNumber (CORBA::ULong newScan)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  for (Workers::iterator wit = workers().begin();
       wit != workers().end();
       ++wit) {
    (*wit)->setScanNumber(newScan);
  }
}

CORBA::ULong SctApiServer::getRunNumber ()
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  CORBA::ULong firstRunNumber=0; // the initialisation value is non necessary or special, it was only added to remove a (misplaced) compiler warning.
  bool first=true;
  for (Workers::iterator wit = workers().begin();
       wit != workers().end();
       ++wit) {
    if (first) {
      first=false;
      firstRunNumber = (*wit)->getRunNumber();
    } else {
      const CORBA::ULong otherRunNumber = (*wit)->getRunNumber();
      if (firstRunNumber != otherRunNumber) {
	std::ostringstream oss;
	oss << "Inconsistent runNumbers " << firstRunNumber
	    << " and " << otherRunNumber
	    << " were reported during SctApiServer::getRunNumber().";
	throw Sct_SctApi::SctApiException(oss.str().c_str());
      }
    }
  }
  if (first) {
    throw Sct_SctApi::SctApiException("Cannot return runNumber from SctApiServer::getRunNumber as there are no crate controllers!");
  }
  return firstRunNumber;
}

/// If all crate controllers agree, getScanNumber() returns the scanNumber they hold, otherwise it throws an Sct_SctApi::SctApiException with 
CORBA::ULong SctApiServer::getScanNumber ()
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  CORBA::ULong firstScanNumber=0; // the initialisation value is non necessary or special, it was only added to remove a (misplaced) compiler warning.
  bool first=true;
  for (Workers::iterator wit = workers().begin();
       wit != workers().end();
       ++wit) {
    if (first) {
      first=false;
      firstScanNumber = (*wit)->getScanNumber();
    } else {
      const CORBA::ULong otherScanNumber = (*wit)->getScanNumber();
      if (firstScanNumber != otherScanNumber) {
	std::ostringstream oss;
	oss << "Inconsistent scanNumbers " << firstScanNumber
	    << " and " << otherScanNumber
	    << " were reported during SctApiServer::getScanNumber().";
	throw Sct_SctApi::SctApiException(oss.str().c_str());
      }
    }
  }
  if (first) {
    throw Sct_SctApi::SctApiException("Cannot return scanNumber from SctApiServer::getScanNumber as there are no crate controllers!");
  }
  return firstScanNumber;
}

void SctApiServer::unsetDebugOption (const char * opt)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  for (Workers::iterator wit = workers().begin();
       wit != workers().end();
       ++wit) {
    (*wit)->unsetDebugOption(opt);
  };
}

void SctApiServer::setDebugOption (const char * opt)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  for (Workers::iterator wit = workers().begin();
       wit != workers().end();
       ++wit) {
    (*wit)->setDebugOption(opt);
  }
}

Sct_SctApi::DebugOptionList * SctApiServer::listEnabledDebugOptions ()
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  return worker()->listEnabledDebugOptions();
}

Sct_SctApi::DebugOptionList * SctApiServer::listDebugOptions ()
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  return worker()->listDebugOptions();
}

void SctApiServer::standardRegisterDumpURID (const Sct::Corba::URID& urid)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  m_crateStore.getCrate(urid)->standardRegisterDumpRod(urid.rod);
}

void SctApiServer::standardRegisterDumpAll ()
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  for (Workers::iterator wit = workers().begin();
       wit != workers().end();
       ++wit) {
    (*wit)->standardRegisterDumpAll();
  }
}

sctConf::ModuleList * SctApiServer::getModuleList () 
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  // Due to redundancy loops etc, there may be a bit of 
  // overlap between the module lists returned by
  // each crate controller, so we must UNION the values they return.
  typedef std::set<std::string> SET;
  SET nearlyTheResult;
  
  for (Workers::iterator wit = workers().begin();
       wit != workers().end();
       ++wit) {
    boost::shared_ptr<sctConf::ModuleList> list1( (*wit)->getModuleList() );
    typedef std::list<std::string> SL;
    const SL & list2 = Sct::copyCorbaToList<sctConf::ModuleList, std::string>(list1.get());
    for (SL::const_iterator it = list2.begin();
	 it!=list2.end();
	 ++it) {
      nearlyTheResult.insert(*it); // accomplishes the unionizing!
    }
  }
 
  return Sct::copyStringSetToCorba<sctConf::ModuleList>(nearlyTheResult);
}

CORBA::Boolean SctApiServer::isRODPresent (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  return worker()->isRODPresent(partition, crate, rod);
}

void SctApiServer::loadConfiguration (const char * filename)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  try {
    worker()->loadConfiguration(filename);
  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by loadConfiguration: " << s.detail << endl;
    throw;
  }
}

void SctApiServer::storeModuleConfigurations ()
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  try {
    worker()->storeModuleConfigurations();
  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by storeModuleConfigurations: " << s.detail << endl;
    throw;
  }
}

void SctApiServer::storeBOCConfigurations ()
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  try {
    worker()->storeBOCConfigurations();
  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by storeModuleConfigurations: " << s.detail << endl;
    throw;
  }
}

void SctApiServer::configureAllModules ()
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  try {
    for (Workers::iterator wit = workers().begin();
	 wit != workers().end();
	 ++wit) {
      (*wit)->configureAllModules();
    };
  } catch(Sct_SctApi::SctApiException &s) {
    Sct::MultiMessageDebugStream m(false,true,true);
    m.severity(MRS_WARNING);
    m << "Exception thrown by configureAllModules: " << s.detail;
    throw;
  }
}

sctConf::ABCDModule SctApiServer::retrieveModule (CORBA::ULong mid)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  return worker()->retrieveModule(mid);
}

/* REMOVE 

CORBA::Short SctApiServer::dspBlockDump (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, 
                                         CORBA::Long dspStart, CORBA::Long numWords, CORBA::Long dspNumber)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  return worker()->dspBlockDump(partition, crate, rod,
                                dspStart, numWords, dspNumber);
}

Sct_SctApi::DataBlock * SctApiServer::dspBlockRead (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::Long dspStart, CORBA::Long numWords, CORBA::Long dspNumber)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  return worker()->dspBlockRead(partition, crate, rod,
                                dspStart, numWords, dspNumber);
}

CORBA::Short SctApiServer::dspBlockWrite (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod,
                                          const Sct_SctApi::DataBlock & buffer, CORBA::ULong dspAddress, 
                                          CORBA::Long dspNumber)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  return worker()->dspBlockWrite(partition, crate, rod, 
                                 buffer, dspAddress, dspNumber);
}

CORBA::ULong SctApiServer::dspSingleRead (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, 
                                          CORBA::ULong dspAddr, CORBA::Long dspNumber)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  return worker()->dspSingleRead(partition, crate, rod, 
                                 dspAddr, dspNumber);
}

void SctApiServer::dspSingleWrite (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::ULong dspAddr, CORBA::ULong val, CORBA::Long dspNumber)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->dspSingleWrite(partition, crate, rod, dspAddr, val, dspNumber);
}

Sct_SctApi::DataBlock * SctApiServer::readSlaveDsp (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::Short s, CORBA::ULong add, CORBA::ULong words)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  return worker()->readSlaveDsp(partition, crate, rod, s, add, words);
}

*/

CORBA::ULong SctApiServer::findModuleSerial (const char * sn)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  // CAREFUL ... TOP LEVEL TRANS METHOD ONLY .. DO NOT MULTIPLY ACROSS WORKERS ...
  return worker()->findModuleSerial(sn);
}

char * SctApiServer::convertToString (CORBA::ULong mid)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  // CAREFUL ... TOP LEVEL TRANS METHOD ONLY .. DO NOT MULTIPLY ACROSS WORKERS ...
  return worker()->convertToString(mid);
}

CORBA::ULong SctApiServer::findModule (CORBA::UShort mur, CORBA::UShort number)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  // CAREFUL ... TOP LEVEL TRANS METHOD ONLY .. DO NOT MULTIPLY ACROSS WORKERS ...
  return worker()->findModule(mur, number);
}

CORBA::ULong SctApiServer::findModuleRod (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::ULong channel) 
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  // CAREFUL ... TOP LEVEL TRANS METHOD ONLY .. DO NOT MULTIPLY ACROSS WORKERS ...
  UINT32 mid = 0;
  mid |= (partition & 0xff) << 24;
  mid |= (crate & 0xff) << 16;
  mid |= (rod & 0xff) << 8;
  mid |= (channel & 0xff) << 0;
  return mid;
}

CORBA::ULong SctApiServer::findBarrelModule (CORBA::ULong barrel, CORBA::ULong row, CORBA::Short number)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  // CAREFUL ... TOP LEVEL TRANS METHOD ONLY .. DO NOT MULTIPLY ACROSS WORKERS ... 
  return worker()->findBarrelModule(barrel, row, number);
}

void SctApiServer::convertToBarrelModule (CORBA::ULong mid, CORBA::ULong & barrel, CORBA::ULong & row, CORBA::Short & number)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  // CAREFUL ... TOP LEVEL TRANS METHOD ONLY .. DO NOT MULTIPLY ACROSS WORKERS ... 
  worker()->convertToBarrelModule(mid, barrel, row, number);
}

CORBA::ULong SctApiServer::findEndcapModule (CORBA::Long disk, CORBA::Long ring, CORBA::Long number)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  // CAREFUL ... TOP LEVEL TRANS METHOD ONLY .. DO NOT MULTIPLY ACROSS WORKERS ... 
  return worker()->findEndcapModule(disk, ring, number);
}

void SctApiServer::convertToEndcapModule (CORBA::ULong mid, CORBA::Long & disk, CORBA::ULong & ring, CORBA::ULong & number)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  // CAREFUL ... TOP LEVEL TRANS METHOD ONLY .. DO NOT MULTIPLY ACROSS WORKERS ... 
  worker()->convertToEndcapModule(mid, disk, ring, number);
}

void SctApiServer::getABCDModule (CORBA::ULong mid, Sct_SctApi::BankType bank)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  try {
    worker()->getABCDModule(mid, bank);
  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by getABCDModule: " << s.detail << endl;
    throw;
  }
}

void SctApiServer::getABCDModules (Sct_SctApi::BankType bank)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  try {
    worker()->getABCDModules(bank);
  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by getABCDModules: " << s.detail << endl;
    throw;
  }
}

/* REMOVE
sctConf::ABCDModule SctApiServer::getABCDModuleRaw (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, 
                                                    CORBA::ULong slot, Sct_SctApi::BankType bank)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  try {
    return worker()->getABCDModuleRaw(partition, crate, rod, slot, bank);
  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by getABCDModules: " << s.detail << endl;
    throw;
  }
}
*/

void SctApiServer::copyABCDModule (CORBA::ULong mid, Sct_SctApi::BankType source, const Sct_SctApi::BankList& to){
  try {
    worker()->copyABCDModule (mid, source, to);
  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by copyABCDModule: " << s.detail << endl;
    throw;
  }
}

void SctApiServer::copyABCDModules (Sct_SctApi::BankType source, const Sct_SctApi::BankList& to){
  try {
    worker()->copyABCDModules(source, to);
  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by copyABCDModules: " << s.detail << endl;
    throw;
  }
}

void SctApiServer::setABCDModule (CORBA::ULong mid, const Sct_SctApi::BankList& banks)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  try {
    worker()->setABCDModule(mid, banks);
  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by setABCDModule: " << s.detail << endl;
    throw;
  }
}

void SctApiServer::setABCDModules (const Sct_SctApi::BankList& banks)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  try {
    for (Workers::iterator wit = workers().begin();
	 wit != workers().end();
	 ++wit) {
      (*wit)->setABCDModules(banks);
    }
  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by setABCDModules: " << s.detail << endl;
    throw;
  }
}

void SctApiServer::sendABCDModule (CORBA::ULong mid, Sct_SctApi::BankType bank, CORBA::ULong type)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  try {
    worker()->sendABCDModule(mid, bank, type);
  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by sendABCDModule: " << s.detail << endl;
    throw;
  }
}

void SctApiServer::sendABCDModules (Sct_SctApi::BankType bank)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  try {
    for (Workers::iterator wit = workers().begin();
	 wit != workers().end();
	 ++wit) {
      (*wit)->sendABCDModules(bank);
    }
  } catch(Sct_SctApi::SctApiException &s) {
    cout << "Exception thrown by sendABCDModules: " << s.detail << endl;
    throw;
  }
}

void SctApiServer::modifyABCDMask (CORBA::ULong mid, const Sct_SctApi::DataBlock & mask)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->modifyABCDMask(mid, mask);
}

void SctApiServer::modifyABCDTrims (CORBA::ULong mid, const Sct_SctApi::CharBlock & trims){
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->modifyABCDTrims(mid, trims);
} 

void SctApiServer::modifyABCDRC (CORBA::ULong mid, CORBA::ULong chip, CORBA::UShort index, CORBA::Double p0, CORBA::Double p1, CORBA::Double p2)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  cout << "SctApiServer::modifyABCDRC called" << endl;
  worker()->modifyABCDRC(mid, chip, index, p0, p1, p2);
}

void SctApiServer::modifyABCDVar (CORBA::ULong typ, CORBA::Double var)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->modifyABCDVar(typ, var);
}

void SctApiServer::modifyABCDModuleVar (CORBA::ULong mid, CORBA::ULong typ, CORBA::Double var)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->modifyABCDModuleVar(mid, typ, var);
}

void SctApiServer::modifyABCDChipVar (CORBA::ULong mid, CORBA::ULong chip, CORBA::ULong typ, CORBA::Double var)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->modifyABCDChipVar(mid, chip, typ, var);
}

void SctApiServer::modifyABCDVarROD (CORBA::ULong typ, CORBA::Double var, const Sct_SctApi::BankList& banks)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->modifyABCDVarROD(typ, var, banks);
  for (Workers::iterator wit = workers().begin();
       wit != workers().end();
       ++wit) {
    (*wit)->modifyABCDVarROD(typ, var, banks);
  }
}

void SctApiServer::modifyABCDModuleVarROD (CORBA::ULong mid, CORBA::ULong typ, CORBA::Double var, const Sct_SctApi::BankList& banks)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->modifyABCDModuleVarROD(mid, typ, var, banks);
}

void SctApiServer::modifyBOCParam (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, 
                                   CORBA::ULong channel, CORBA::ULong type, CORBA::ULong val)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->modifyBOCParam(partition, crate, rod, channel, type, val);
}

void SctApiServer::lasersOff ()
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  for (Workers::iterator wit = workers().begin();
       wit != workers().end();
       ++wit) {
    (*wit)->lasersOff();
  }
}

void SctApiServer::timSetFrequency (CORBA::ULong partition, CORBA::ULong crate, 
                                    CORBA::Double trigFreq, CORBA::Double rstFreq)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->timSetFrequency(partition, crate, trigFreq, rstFreq);
}

void SctApiServer::freeTriggers (CORBA::ULong partition, CORBA::ULong crate)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->freeTriggers(partition, crate);
}

void SctApiServer::stopTriggers (CORBA::ULong partition, CORBA::ULong crate)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->stopTriggers(partition, crate);
}

void SctApiServer::timL1A (CORBA::ULong partition, CORBA::ULong crate)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->timL1A(partition, crate);
}

void SctApiServer::timCalL1A (CORBA::ULong partition, CORBA::ULong crate, CORBA::Short delay)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->timCalL1A(partition, crate, delay);
}

CORBA::Short SctApiServer::timReadRegister(CORBA::ULong partition,
                                           CORBA::ULong crate,
                                           CORBA::Short value) 
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  return worker()->timReadRegister(partition, crate, value);
}

void SctApiServer::timWriteRegister (CORBA::ULong partition, CORBA::ULong crate, CORBA::Short reg, CORBA::Short value)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->timWriteRegister(partition, crate, reg, value);
}

void SctApiServer::sendTimTrigger (CORBA::ULong partition,
                                   CORBA::ULong crate,
                                   Sct_SctApi::Trigger_ptr trig) 
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->sendTimTrigger(partition, crate, trig);
}

void SctApiServer::sendTrigger (CORBA::ULong partition,
                                CORBA::ULong crate,
                                CORBA::ULong rod,
                                Sct_SctApi::Trigger_ptr trig)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->sendTrigger(partition, crate, rod, trig);
}

CORBA::Boolean SctApiServer::hasFinishedScanning() {
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  return m_assm->scannersHaveFinished();
};

void SctApiServer::abortScan() {
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  for (Workers::iterator wit = workers().begin();
       wit != workers().end();
       ++wit) {
    (*wit)->abortScan(); // it does not hurt to send abortScan to those CCs whose scans have already stopped
  };
};

void SctApiServer::doScan(Sct_SctApi::Scan_ptr thisScan)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  mrsStream << "SCTAPISERVER" << MRS_INFORMATION << MRS_QUALIF("SctApiServer") << MRS_TEXT("Starting cooked scan") << ENDM;

  try {
    // OLD worker()->doScan(thisScan);
    // NEW START
    {
      for (Workers::iterator wit = workers().begin();
	   wit != workers().end();
	   ++wit) {
	(*wit)->doScan(thisScan); // if these lines are ever run in parallel, make sure that SctApiServer::doScan does not return until every m_assm->monotorThisScanner(w) call has returned!
	m_assm->monitorThisScanner(*wit);
      };
    };
    // NEW STOP
  } catch(Sct_SctApi::SctApiException &s) {

    std::ostringstream os;
    os << "Scan aborting as exception was thrown by doScan: [" << s.detail << "]";

    cout << os.str() << endl;
    mrsStream << "COOKED_SCAN_ABORTING" << MRS_ERROR << MRS_QUALIF("SctApiServer") << MRS_TEXT(os.str()) << ENDM;

    // Do our best to clear things up and bring SctApiServer back to "normal" non-scanning state
    try {
      this->abortScan();
    } catch(...) {
      Sct::MultiMessageDebugStream m;
      m << "Failed to clean things up after exception was thrown by doScan at " << __LINE__ << " in " << __FILE__;
      m.flush();
      // do not throw, so that users can see cause original exception
    };

    throw;  // rethrow original exception that caused the problem
  }
}

void SctApiServer::doRawScan (Sct_SctApi::Scan_ptr thisScan,
                              CORBA::Long delay,
                              CORBA::Long width,
                              CORBA::Boolean configureModules,
                              CORBA::Boolean clockByTwo)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  mrsStream << "SCTAPISERVER" << MRS_INFORMATION << MRS_QUALIF("SctApiServer") << MRS_TEXT("Starting raw scan") << ENDM;

  try {
    // OLD worker()->doScan(thisScan);
    // NEW START
    {
      for (Workers::iterator wit = workers().begin();
	   wit != workers().end();
	   ++wit) {
	(*wit)->doRawScan(thisScan,delay,width,configureModules,clockByTwo); // if these lines are ever run in parallel, make sure that SctApiServer::doScan does not return until every m_assm->monotorThisScanner(w) call has returned!
	m_assm->monitorThisScanner(*wit);
      };
    };
    // NEW STOP
  } catch(Sct_SctApi::SctApiException &s) {

    std::ostringstream os;
    os << "Scan aborting as exception was thrown by doRawScan: [" << s.detail << "]";

    cout << os.str() << endl;
    mrsStream << "RAW_SCAN_ABORTING" << MRS_ERROR << MRS_QUALIF("SctApiServer") << MRS_TEXT(os.str()) << ENDM;

    // Do our best to clear things up and bring SctApiServer back to "normal" non-scanning state
    try {
      this->abortScan();
    } catch(...) {
      Sct::MultiMessageDebugStream m;
      m << "Failed to clean things up after exception was thrown by doRawScan at " << __LINE__ << " in " << __FILE__;
      m.flush();
      // do not throw, so that users can see cause original exception
    };

    throw;  // rethrow original exception that caused the problem
  }
}

/* REMOVE
void SctApiServer::printBOCSetup (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->printBOCSetup(partition, crate, rod);
}
*/

Sct_SctApi::BOCChannelList * SctApiServer::currentBOCSetup (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod) 
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  return worker()->currentBOCSetup(partition, crate, rod);
}

sctConf::BOCGlobalConfig SctApiServer::currentBOCRegisters (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod) 
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  return worker()->currentBOCRegisters(partition, crate, rod);
}

void SctApiServer::printABCDModule (CORBA::ULong mid)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->printABCDModule(mid);
}

void SctApiServer::printABCDRodModule (CORBA::ULong mid, Sct_SctApi::BankType bank)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->printABCDRodModule(mid, bank);
}

void SctApiServer::decodeEvent (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::Short sl, CORBA::Short index, CORBA::Boolean extFlag, CORBA::Boolean errorType)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->decodeEvent(partition, crate, rod, sl, index, extFlag, errorType);
}

void SctApiServer::rawData (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::Long delay, CORBA::Long units, CORBA::Boolean setMask)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->rawData(partition, crate, rod, delay, units, setMask);
}

void SctApiServer::rawDataWithTrigger (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::Long delay, CORBA::Long units, CORBA::Boolean setMask, Sct_SctApi::Trigger_ptr trig)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->rawDataWithTrigger(partition, crate, rod, delay, units, setMask, trig);
}

void SctApiServer::sendL1A (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->sendL1A(partition, crate, rod);
}

Sct_SctApi::CharBlock * SctApiServer::probe (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  return worker()->probe(partition, crate, rod);
}

Sct_SctApi::CharBlock * SctApiServer::probeHarness (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::ULong harness)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  return worker()->probeHarness(partition, crate, rod, harness);
}

Sct_SctApi::CharBlock * SctApiServer::probeWithTrigger (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, Sct_SctApi::Trigger_ptr trig)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  return worker()->probeWithTrigger(partition, crate, rod, trig);
}

Sct_SctApi::CharBlock * SctApiServer::probeHarnessWithTrigger (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, Sct_SctApi::Trigger_ptr trig, CORBA::ULong harness)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  return worker()->probeHarnessWithTrigger(partition, crate, rod, trig, harness);
}

Sct_SctApi::CharMatrix * SctApiServer::probeScanHarnessWithTrigger (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, Sct_SctApi::Scan_ptr scan, CORBA::ULong harness)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  return worker()->probeScanHarnessWithTrigger(partition, crate, rod, scan, harness);
}

CORBA::Boolean SctApiServer::checkAllModulesProbe (const char * values)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  return worker()->checkAllModulesProbe(values);
}

Sct_SctApi::AutoResultList * SctApiServer::autoConfigureScan ()
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  return worker()->autoConfigureScan();
}

CORBA::ULong SctApiServer::getRodSlot (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  return worker()->getRodSlot(partition, crate, rod);
}

void SctApiServer::testLinkOutSelect (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::ULong link)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  worker()->testLinkOutSelect(partition, crate, rod, link);
}

CORBA::ULong SctApiServer::testLinkOutRetrieve (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  return worker()->testLinkOutRetrieve(partition, crate, rod);
}

Sct_SctApi::Scan_ptr SctApiServer::createScan ()
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  std::cout << "Asked to create a scan\n";

  SharedScanServer defScan = ::Sct::DestroyingDeleter<ScanServer>::shared_ptr( new ScanServer ); // , *defaultScan);

  scans.push_back(defScan);

  return defScan->_this();
}

Sct_SctApi::Scan_ptr SctApiServer::getDefaultScan ()
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  std::cout << "Asked for a reference to the default scan\n";

  return defaultScan->_this();
}

Sct_SctApi::Trigger_ptr SctApiServer::createTrigger ()
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  std::cout << "Asked to create a trigger\n";

  TriggerServer * defTrig = new TriggerServer;
#warning DOES ANYONE EVER DELETE THIS? I THINK NOT! (ChrisLester)
  
  return defTrig->_this();
}

/*
Sct_SctApi::Trigger_ptr SctApiServer::getDefaultTrigger ()
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  std::cout << "Asked for the default trigger\n";
#warning I DONT THINK ANYONE ACTUALLY SETS THE DEFAULT TRIGGER! THIS IS AN INVITATION TO SEG-FAULT! (ChrisLester)
  std::cout << "I will now do my best to segmentation fault." << std::endl;
  assert(defaultTrigger);
  return defaultTrigger->_this();
}
*/

Sct_SctApi::ScanList * SctApiServer::knownScans ()
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  std::cout << "In known scans\n";

  Sct_SctApi::ScanList * myScans = SequenceUtils::newSafeSequence<Sct_SctApi::ScanList>(scans.size());

  std::cout << "Copying sequence of length " << scans.size() << std::endl;

  unsigned int i=0;
  for(Scans::iterator it=scans.begin();
      it != scans.end();
      (++i,++it)) {
    ScanServer & thisScan = **it;
    (*myScans)[i] = thisScan._this();
  }

  return myScans;
}

Sct_SctApi::CrateIPC_ptr SctApiServer::getCrate(const Sct::Corba::UCID& ucid)
{
  return Sct_SctApi::CrateIPC::_duplicate( m_crateStore.getCrate(ucid) );
}

Sct_SctApi::CrateIPC_ptr SctApiServer::getCrateLegacy(CORBA::ULong partition, CORBA::ULong crate)
{
  Sct::Corba::UCID ucid;
  ucid.partition=partition;
  ucid.crate=crate;
  return this->getCrate(ucid);
}

Sct::Corba::UCIDs* SctApiServer::getCrateUCIDs() {
  const CrateStore::UCIDList & ucidList = m_crateStore.ucidList(); 
  Sct::Corba::UCIDs * ret = SequenceUtils::newSafeSequence<Sct::Corba::UCIDs>(ucidList.size());
  unsigned int i=0;
  for (CrateStore::UCIDList::const_iterator it = ucidList.begin();
       it != ucidList.end();
       (++it, ++i)) {
    (*ret)[i] = Sct::copyUCIDToCorba(*it);
  };
  return ret;
};

/// See the note which deprecates this method in the header file!
Sct_SctApi::CrateIPC_var SctApiServer::worker() 
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  if (m_crateStore.workers().empty()) {
    Sct_SctApi::CrateIPC_var ret; // will be nil!  Good!
    return ret;
  } else {
    return m_crateStore.workers().front(); // yes this is indeed arbitrary! It is only condoned as this method will be removed shortly.
  };
}

SctApiServer::Workers & SctApiServer::workers() 
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
  return m_crateStore.workers();
}

void SctApiServer::shutdown ()
{
  std::cout << "MOO MOO MOO " << __FILE__ << " " << __LINE__ << " " << __PRETTY_FUNCTION__ << std::endl;
#warning WHAT SHOULD SHUTDOWN DO IN SCTAPISERVER.  MAYBE IT SHOULD CALL SHUTDOWNALL ??
  this->stop();
};
