#ifndef WORKERGROUP_H
#define WORKERGROUP_H
#include <list>
#include <vector>
#include <algorithm>

#include <boost/thread.hpp>
#include <boost/timer.hpp>

#include "Sct/LogicErrors.h"
#include "Sct/ISUtilities.h"
#include "Sct/SctNames.h"
#include "IS/ServiceStatus.h"

namespace SctService{

/**
  Control part of the observer/worker pattern.
  Contains the queue to which objects of type T are pushed.
  Also useful functions to control and query the worker threads.
  The template class ``T'' is the data/instructions which the workers at to work on.
  Derive this class to implement/modify the behaviour of the worker threads.  At the least,
  the work method needs an implementation.
  T must be DefaultConstructible
  For a simple example, see the test/MyWorkerGroup.h file
*/
template <class T>
class WorkerGroup : private boost::noncopyable {
public:
    typedef T DataType;
    
    /// Start this worker group. @param the number of worker threads to be started
    void go(unsigned nWorker=1);    

    /**
       get the number of objects in the queue
       @note for information only, not to be used in thread-control.
    */
    unsigned queueSize() const;
    /**
     * get the number of workers in the map
     */
    unsigned nWorkers() const;
    /**
       get the number of jobs done
     */
    unsigned long nDone() const;
    /**
     * get the number of threads which are busy.
     * @note for information only, not to be used in thread-control.
     */
    unsigned busy() const;
    /**
     * Raise flag to stop all worker threads.
     */
    void stop();
    /*
     * Raise flag to pause all worker threads.
     * @param true=pause, false=unpause/play/restart 
     */
    void setPaused(bool paused=true);
    /**
     * Call this to make sure that Worker threads continue to completion.
     */
    void join();
    /**
     * get the number of paused threads
     */
    unsigned paused() const;
    /**
     * Set sleep time for all workers, if no items are on the queue.
     * @param sleep time in msec
     */
    void setSleepTime(unsigned msec);
    /**
      Get the sleep time
      */
    unsigned getSleepTime() const {return m_sleeptime;}
    /*
     * Pop an object off the front of the queue. @return T to the object
     * if the queue isn't empty, otherwise, T()
     */
    DataType pop();
    /**
     * Push an object onto the back of the queue.
     */
    void push(DataType ob);    

    /*
      Set First In First Out (if true) or First In Last Out (if false).
      FILO is more performant, since less disk swap is necessary, 
      but a bit less intuitive for the user!
    */
    void setFifo(bool val);
    /*
      get FIFO (true) or FILO (false) state.
    */
    bool isFifo() const;
    /// destructor
    virtual ~WorkerGroup() {;} 
    
    ///Constructor
    WorkerGroup() : m_run(true), m_pause(false), m_sleeptime(1), m_reportInterval(1000), 
      m_nDone(0), m_fifo(true) {}

    /// set name to report to IS as
    void reportTo(std::string isname);
    /// set interval in msec between reports
    void reportInterval(int msec);
protected:
    ///Inner class to hold thread specific stuff
    class Worker {
    public:
	Worker(WorkerGroup<DataType>& group) : group(group), m_busy(false), m_paused(false) {
	}
	
	Worker(const Worker& w) : group(w.group), m_busy(false), m_paused(false) {}

	void operator() () {
	    group.threadMain(*this);
	}
	WorkerGroup<DataType>& group;
	/// Safely get busy status
	bool busy();
	/// Safely get paused status
	bool paused();
	/// Safely set busy status
	void setBusy(bool v);
	/// Safely set paused status
	void setPaused(bool v);
	/// Copy without trying to copy mutex!
	//Worker& operator=(const Worker& w)
    private:
	boost::mutex m_statusMutex;      ///< permits status to be locked
	volatile bool m_busy;            ///< true when actively working
	volatile bool m_paused;          ///<true if paused
    };

    /// holds data used by the reporting thread
    class Reporter {
    public:
      Reporter(WorkerGroup<DataType>& group) : group(group){ 
      }
      void operator() (){
	group.reportThreadMain(*this);
      }
      WorkerGroup<DataType>& group;
      ServiceStatus last;
    };

    /** thread which reports the progress of the workers to IS */
    virtual void reportThreadMain(Reporter& reporter) throw();
    
    /**
      The threadMain method.  Can be overriden if sub-classes want to completly change the
      behaviour!  Current implementation checks various flags for pausing, stopping etc,
      then calls popAndWork.  If popAndWork returns false, then it calls waitForData
      */
    virtual void threadMain(Worker& worker) throw();
    /**
      If the queue length != 0, removes an element from the queue and class work
      @returns false if the queue length is 0.
      */
    virtual bool popAndWork(Worker& worker);
    /**
      Called if there is no data on the queue.  Calls wait in default implementation.
      */
    virtual void waitForData();
    /**
      To be overriden!  Should actually do something with the data!
      */
    virtual void work(DataType data) throw() = 0;
    /**
      Causes this thread to sleep for msec milliseconds.
      */
    void wait(int msec);
    /**
      threadMain should call this with its parameter.
      */
    void addWorker(Worker& worker);
    
    typedef typename std::vector<Worker*> WorkerStore;
    typedef typename WorkerStore::iterator Iterator;
    typedef typename WorkerStore::const_iterator ConstIterator;
    
    /**
     * @name Worker thread information @{ 
     * This map allows each thread to access different particular 
     * information from its own Worker.
     */
     WorkerStore m_workers;

    /**
     * @name Queue @{ the queue itself contains pointers to objects to be
     * worked on.
     */
    std::list<DataType> m_queue;
    

    /// boost thread group from which threads can be created
    boost::thread_group m_thread_group;
    
    volatile bool m_run;  ///< if false, threads are to stop.
    volatile bool m_pause; ///< does not execute work() if true
    volatile unsigned m_sleeptime; ///< time to sleep if queue is empty	
    std::string m_reportTo; ///<name to report to IS
    volatile int m_reportInterval;   ///<name to report to IS
    
    mutable boost::mutex m_queueAccess;  ///< queueLock
    mutable boost::mutex m_vectorAccess; ///< vectorOfWorkersLock    
    mutable boost::recursive_mutex m_reportAccess; ///< vectorOfWorkersLock    
    volatile unsigned long m_nDone;   ///< count of number of jobs done
    volatile double m_time;   ///< time taken working
    friend class Worker;
    friend class Reporter;
    void addTime(double t);            ///< timer watch
    bool m_fifo;                       ///< FIFO if true, otherwise FILO
    /** 
	gets from front or back of queue (no locking) depending on m_fifo
    */
    DataType unlockedPopFrontOrBack();
};


template <class T>
inline void WorkerGroup<T>::go(unsigned nWorker){    
    // create nWorker new threads
    for (unsigned i=0; i<nWorker; ++i) {
	m_thread_group.create_thread(Worker(*this));
    }
    m_thread_group.create_thread(Reporter(*this));
}

template <class T>
inline void WorkerGroup<T>::threadMain(Worker& worker) throw() { 
    //As the worker is owned by some boost thing, we just need to add it.
    addWorker(worker);
    
    while (m_run) {                        // if !m_run, thread will finish.
	while (m_pause && m_run) {     // while paused, and not stopped, wait in sleep loop.
	    worker.setPaused(true);
	    wait(m_sleeptime);
	}
	worker.setPaused(false);
	if (!m_run) return;                     // if no longer running, stop.
	
	if (!popAndWork(worker)) {
	    waitForData();    // if queue is empty, sleep.
	}
    }
    return;
}

template <class T>
inline void WorkerGroup<T>::reportThreadMain(Reporter& reporter) throw() { 
  bool problemsToMrs=false;
  bool first=true;
  std::cout << "Status to: [" << m_reportTo << "]" << std::endl;
  while (m_run) {
    while (m_pause && m_run) {
      wait(m_reportInterval);
    }
    if (!m_run) return; 
    wait(m_reportInterval);
    {
      boost::recursive_mutex::scoped_lock lock(m_reportAccess);
      if (!lock.locked()) {
	continue;
      }
      bool modified = (reporter.last.queue != queueSize()||
		       reporter.last.done != nDone() || first);
      if (first) first=false;
      if (modified){
	reporter.last.averageTimeEach=0.;
	reporter.last.timeLeft=0.;
	reporter.last.queue = queueSize();
	reporter.last.done  = nDone();
	if (reporter.last.done!=0) {
	  reporter.last.averageTimeEach = static_cast<float>(m_time)/static_cast<float>(reporter.last.done);
	}
	unsigned nw=nWorkers();
	if (nw!=0) {
	  reporter.last.timeLeft = static_cast<float>(reporter.last.queue)*static_cast<float>(reporter.last.averageTimeEach) / static_cast<float>(nw);
	}
	if (m_reportTo!="") {
	  try {Sct::ISUtilities::addOrUpdateOrThrow(m_reportTo, reporter.last,
						    __FILE__, __LINE__, MRS_DIAGNOSTIC);
	  }catch(Sct::Throwable& e){
	    if (!problemsToMrs){
	      problemsToMrs=true;
	      e.sendToMrs(MRS_DIAGNOSTIC);
	    }
	  }
	}
      }
    }
  }
  return;
}

template<class T>
inline void WorkerGroup<T>::addWorker(Worker& worker) {
    boost::mutex::scoped_lock lock(m_vectorAccess);
    m_workers.push_back(&worker);
}

template <class T>
inline void WorkerGroup<T>::push(DataType ob){
    boost::mutex::scoped_lock lock(m_queueAccess);
    if (lock.locked()) m_queue.push_back(ob);
    else throw Sct::IllegalStateError("Couldn't get queue lock", __FILE__, __LINE__);
}

template <class T>
inline T WorkerGroup<T>::unlockedPopFrontOrBack(){
  if (m_queue.empty()) throw Sct::IllegalStateError("Queue is emtpy!", __FILE__, __LINE__);
  if (m_fifo){
    DataType data=m_queue.front();
    m_queue.pop_front();
    return data;
  }else{
    DataType data=m_queue.back();
    m_queue.pop_back();
    return data;
  }
}

template <class T>
inline T WorkerGroup<T>::pop(){
    boost::mutex::scoped_lock lock(m_queueAccess);
    if (lock.locked()) {
      return unlockedPopFrontOrBack();
    } else throw Sct::IllegalStateError("Couldn't get queue lock", __FILE__, __LINE__);
}

template <class W>
inline bool WorkerGroup<W>::popAndWork(Worker& worker){
    boost::mutex::scoped_lock lock(m_queueAccess);
    if (lock.locked()) {	
	if (!m_queue.empty()) {
            DataType data=unlockedPopFrontOrBack();
	    lock.unlock();  		 //Don't hold onto the lock for too long!
	    worker.setBusy(true);        // set the busy flag
	    boost::timer t;
	    work(data);                  // call the function to do work.
	    addTime(t.elapsed());
	    worker.setBusy(false);       // unset the busy flag 
	    return true;
	} else return false;
    } else throw Sct::IllegalStateError("Couldn't get queue lock", __FILE__, __LINE__);
}

template <class W>    
inline unsigned WorkerGroup<W>::queueSize() const{
    return m_queue.size();
}

template <class W>    
inline unsigned long WorkerGroup<W>::nDone() const{
    return m_nDone;
}

template <class W>
inline unsigned WorkerGroup<W>::nWorkers() const{
    return m_workers.size();
}

template <class W>
inline void WorkerGroup<W>::stop(){
    m_run = false;
    join();
}

template <class W>
inline void WorkerGroup<W>::join(){
    m_thread_group.join_all();
}

template <class W>
inline void WorkerGroup<W>::setPaused(bool p){
    m_pause = p;
}

template <class W>
inline unsigned WorkerGroup<W>::paused() const{
    unsigned npaused=0;
    for (ConstIterator it=m_workers.begin(); it!=m_workers.end(); ++it) 
	if ( (*it)->paused() ) ++npaused;

    return npaused;
}

template <class W>
inline unsigned WorkerGroup<W>::busy() const{
    unsigned nbusy=0;
    for (ConstIterator it=m_workers.begin(); it!=m_workers.end(); ++it) {
	if ( (*it)->busy() ) ++nbusy;
    }
    return nbusy;
}

template <class W>
inline void WorkerGroup<W>::setSleepTime(unsigned t){
    m_sleeptime = t;
}

template <class W>
inline void WorkerGroup<W>::wait(int wait_time) {
    boost::xtime xt;
    boost::xtime_get(&xt, boost::TIME_UTC);
    xt.nsec += wait_time*1000000;
    boost::thread::sleep(xt);
}

template <class W>
inline void WorkerGroup<W>::waitForData(){
    wait(getSleepTime());
}

template <class T>
void WorkerGroup<T>::reportTo(std::string isname){
  boost::recursive_mutex::scoped_lock lock(m_reportAccess);
  m_reportTo=isname;
}

template <class T>
inline void WorkerGroup<T>::reportInterval(int msec){
  boost::recursive_mutex::scoped_lock lock(m_reportAccess);
  if (msec>100){
    m_reportInterval=msec;
  }
}

template <class W>
inline void WorkerGroup<W>::addTime(double t){
  boost::recursive_mutex::scoped_lock lock(m_reportAccess);
  m_time+=t;
  ++m_nDone;
}

template <class W>
inline void WorkerGroup<W>::setFifo(bool val){
  boost::mutex::scoped_lock lock(m_queueAccess);
  if (!lock.locked()) throw Sct::IllegalStateError("Could not get status lock",__FILE__, __LINE__);
  m_fifo=val;
}

template <class W>
inline bool WorkerGroup<W>::isFifo()const{
  boost::mutex::scoped_lock lock(m_queueAccess);
  if (!lock.locked()) throw Sct::IllegalStateError("Could not get status lock",__FILE__, __LINE__);
  return m_fifo;
}



template <class W>
inline bool WorkerGroup<W>::Worker::busy(){
  boost::mutex::scoped_lock lock(m_statusMutex);
  if (!lock.locked()) throw Sct::IllegalStateError("Could not get status lock",__FILE__, __LINE__);
  return m_busy;
}

template <class W>
inline bool WorkerGroup<W>::Worker::paused(){
  boost::mutex::scoped_lock lock(m_statusMutex);
  if (!lock.locked()) throw Sct::IllegalStateError("Could not get status lock",__FILE__, __LINE__);
  return m_paused;
}

template <class W>
void WorkerGroup<W>::Worker::setBusy(bool val){
  boost::mutex::scoped_lock lock(m_statusMutex);
  if (!lock.locked()) throw Sct::IllegalStateError("Could not get status lock",__FILE__, __LINE__);
  m_busy=val;
}

template <class W>
inline void WorkerGroup<W>::Worker::setPaused(bool val){
  boost::mutex::scoped_lock lock(m_statusMutex);
  if (!lock.locked()) throw Sct::IllegalStateError("Could not get status lock",__FILE__, __LINE__);
  m_paused=val;
}

} // end of namepace SctService
#endif //#ifndef WORKERGROUP_H
