#include "marshalling.h"

ABCDModule copyCorbaToABCDModule(const sctConf::ABCDModule& corbaModule) {
  ///This fails if sctStructure has been updated without updating the definition of ABCDModule in configuration.idl
  assert(sizeof(ABCDModule) == sizeof(sctConf::ABCDModule));

  // There are unused octets in ABCDModule, and these are not 
  // always zeroed in the corba representation, and there is suspicion growing
  // that the rod may be paying unwanted attention to these unused bits.
  // Whether or not the ROD is paying attention, it is still annoying
  // to have random bits floating around in dumps of PrimLists when you
  // are trying to debug differences between prim lists in different versions
  // of the software, so here is an experimental bit-zeroing field by field
  // copy to try to remove this problem:

  const bool zeroUnaddressableBits = true;
  if (zeroUnaddressableBits) {
    
    ABCDModule result;

    // first zero everything:
    for(unsigned int i=0; i<sizeof(result); i++) {
      ((char*)&result)[i] = 0;
    }

    // now copy the addressable structures through:
    for (unsigned int i=0; i<N_SCT_CHIPS; ++i) {
      ABCDChip & resultChip = result.chip[i];
      const sctConf::ABCDChip & corbaChip = corbaModule.chip[i]; 
      
      resultChip.active = corbaChip.active;
      resultChip.address = corbaChip.address;
      resultChip.unused = corbaChip.unused;
      resultChip.target = corbaChip.target;
      // Do ABCDBasic FIELD-BY-FIELD as it has unaddressed
      // bytes hidden inside it and we want them to be zeroed! :
      {
        resultChip.basic.config = reinterpret_cast<const ABCDConfig&>(corbaChip.basic.config);
        resultChip.basic.unused = corbaChip.basic.unused;
        resultChip.basic.vthr = corbaChip.basic.vthr;
        resultChip.basic.vcal = corbaChip.basic.vcal;
        resultChip.basic.delay = corbaChip.basic.delay;
        resultChip.basic.preamp = corbaChip.basic.preamp;
        resultChip.basic.shaper = corbaChip.basic.shaper;
        resultChip.basic.mask[0] = corbaChip.basic.mask[0];
        resultChip.basic.mask[1] = corbaChip.basic.mask[1];
        resultChip.basic.mask[2] = corbaChip.basic.mask[2];
        resultChip.basic.mask[3] = corbaChip.basic.mask[3];
      };
      // caldata:
      {
        resultChip.caldata.rc_function = corbaChip.caldata.rc_function;
        resultChip.caldata.rc_params[0] = corbaChip.caldata.rc_params[0];
        resultChip.caldata.rc_params[1] = corbaChip.caldata.rc_params[1];
        resultChip.caldata.rc_params[2] = corbaChip.caldata.rc_params[2];
        resultChip.caldata.c_factor = corbaChip.caldata.c_factor;
        resultChip.caldata.unused[0] = corbaChip.caldata.unused[0];
        resultChip.caldata.unused[1] = corbaChip.caldata.unused[1];
        resultChip.caldata.unused[2] = corbaChip.caldata.unused[2];
      };
      for (unsigned int chan=0; chan<N_SCT_CHANS; ++chan) {
        resultChip.trim[chan] = corbaChip.trim[chan];
      };
    };
    result.present = corbaModule.present;
    result.active = corbaModule.active;
    result.select = corbaModule.select;
    result.groupId = corbaModule.groupId;
    result.bocConfig = reinterpret_cast<BOCConfig*>(corbaModule.bocConfig);
    result.pTTC = corbaModule.pTTC;
    result.rTTC = corbaModule.rTTC;
    result.rx[0] = corbaModule.rx[0];
    result.rx[1] = corbaModule.rx[1];
    result.unused1[0] = corbaModule.unused1[0];
    result.unused1[1] = corbaModule.unused1[1];
    result.unused1[2] = corbaModule.unused1[2];
    result.unused1[3] = corbaModule.unused1[3];
    result.unused2[0] = corbaModule.unused2[0];
    result.unused2[1] = corbaModule.unused2[1];
    result.unused2[2] = corbaModule.unused2[2];
    result.unused2[3] = corbaModule.unused2[3];
    
    return result;
  } else {
    return ABCDModule(reinterpret_cast<const ABCDModule&>(corbaModule));
  };
}

sctConf::ABCDModule copyABCDModuleToCorba(const ABCDModule &module) {

  ///This fails if sctStructure has been updated without updating the definition of ABCDModule in configuration.idl
  assert(sizeof(ABCDModule) == sizeof(sctConf::ABCDModule));
  return sctConf::ABCDModule(reinterpret_cast<const sctConf::ABCDModule&>(module)) ;
}

SctConfiguration::SlaveConfig copyCorbaToSlaveConfig(const sctConf::SlaveConfig& corbaConfig) {
  SctConfiguration::SlaveConfig realConfig;

  realConfig.ipramFile = corbaConfig.ipramFile;
  realConfig.idramFile = corbaConfig.idramFile;
  realConfig.extFile = corbaConfig.extFile;

  return realConfig;
}

SctConfiguration::RodConfig copyCorbaToRodConfig(const sctConf::RodConfig& corbaConfig) {
  SctConfiguration::RodConfig realConfig;

  realConfig.baseAddress = corbaConfig.baseAddress;
/*   realConfig.mapSize = corbaConfig.mapSize; */
  realConfig.resetLevel = corbaConfig.resetLevel;
/*   realConfig.numSlaves = corbaConfig.numSlaves; */

  if(corbaConfig.slaves.length() != 4) {
    throw SctConfiguration::ConfigurationException("RodConfig must contain 4 slaves!");
  }

  for(int i=0; i<4; i++) {
    realConfig.slaves[i] = copyCorbaToSlaveConfig(corbaConfig.slaves[i]);
  }

  return realConfig;
}

sctConf::SlaveConfig* copySlaveConfigToCorba(const SctConfiguration::SlaveConfig &config) {
  sctConf::SlaveConfig* result = new sctConf::SlaveConfig;

  result->ipramFile = CORBA::string_dup(config.ipramFile.c_str());
  result->idramFile = CORBA::string_dup(config.idramFile.c_str());
  result->extFile = CORBA::string_dup(config.extFile.c_str());

  /* AJB replaced mallocs*/

  return result;
}

sctConf::RodConfig* copyRodConfigToCorba(const SctConfiguration::RodConfig &config) {
  sctConf::RodConfig *result = new sctConf::RodConfig;

  result->baseAddress = config.baseAddress;
  result->resetLevel = config.resetLevel;
/*   result->mapSize = config.mapSize; */
/*   result->numSlaves = config.numSlaves; */

  result->slaves = sctConf::SlaveList(4);
  result->slaves.length(4);

  for(int i=0; i<4; i++) {
    sctConf::SlaveConfig& slave = result->slaves[i];

    const SctConfiguration::SlaveConfig &conf = config.slaves[i];

    slave.ipramFile = CORBA::string_dup(conf.ipramFile.c_str());
    slave.idramFile = CORBA::string_dup(conf.idramFile.c_str());
    slave.extFile = CORBA::string_dup(conf.extFile.c_str());

    /*AJB replaced mallocs */
  }

  return result;
}

SctConfiguration::BOCChannelConfig copyCorbaToBOCChannelConfig(const sctConf::BOCChannelConfig& corbaConfig) {
  SctConfiguration::BOCChannelConfig newConf;

  newConf.current = corbaConfig.current;
  newConf.delay = corbaConfig.delay;
  newConf.markSpace = corbaConfig.markSpace;

  newConf.threshold0 = corbaConfig.threshold0;
  newConf.delay0 = corbaConfig.delay0;

  newConf.threshold1 = corbaConfig.threshold1;
  newConf.delay1 = corbaConfig.delay1;

  return newConf;
}

sctConf::BOCChannelConfig* copyBOCChannelConfigToCorba(const SctConfiguration::BOCChannelConfig &config) {
  sctConf::BOCChannelConfig *res = new sctConf::BOCChannelConfig;

  res->current = config.current;
  res->delay = config.delay;
  res->markSpace = config.markSpace;

  res->threshold0 = config.threshold0;
  res->delay0 = config.delay0;

  res->threshold1 = config.threshold1;
  res->delay1 = config.delay1;

  return res;
}

SctConfiguration::BOCGlobalConfig copyCorbaToBOCGlobalConfig(const sctConf::BOCGlobalConfig& corbaConfig) {
  SctConfiguration::BOCGlobalConfig newConf;

  newConf.validMask = corbaConfig.validMask;
  newConf.clockControl = corbaConfig.clockControl;
  newConf.rxDataMode = corbaConfig.rxDataMode;
  newConf.rxDacClear = corbaConfig.rxDacClear;
  newConf.txDacClear = corbaConfig.txDacClear;
  newConf.vernierFinePhase = corbaConfig.vernierFinePhase;
  newConf.vernierClockPhase0 = corbaConfig.vernierClockPhase0;
  newConf.vernierClockPhase1 = corbaConfig.vernierClockPhase1;
  newConf.bpmClockPhase = corbaConfig.bpmClockPhase;
  newConf.bregClockPhase = corbaConfig.bregClockPhase;

  return newConf;
}

sctConf::BOCGlobalConfig copyBOCGlobalConfigToCorba(const SctConfiguration::BOCGlobalConfig &config) {
  sctConf::BOCGlobalConfig res;

  res.validMask = config.validMask;
  res.clockControl = config.clockControl;
  res.rxDataMode = config.rxDataMode;
  res.rxDacClear = config.rxDacClear;
  res.txDacClear = config.txDacClear;
  res.vernierFinePhase = config.vernierFinePhase;
  res.vernierClockPhase0 = config.vernierClockPhase0;
  res.vernierClockPhase1 = config.vernierClockPhase1;
  res.bpmClockPhase = config.bpmClockPhase;
  res.bregClockPhase = config.bregClockPhase;

  return res;
}

SctConfiguration::TimConfig copyCorbaToTimConfig(const sctConf::TimConfig& corbaConfig) {
  SctConfiguration::TimConfig result;

  result.baseAddress = corbaConfig.baseAddress;
/*   result.mapSize = corbaConfig.mapSize; */
  result.trigFrequency = corbaConfig.trigFrequency;
  result.resetFrequency = corbaConfig.resetFrequency;

  return result;
}

sctConf::TimConfig copyTimConfigToCorba(const SctConfiguration::TimConfig &config) {
  sctConf::TimConfig res;
  res.baseAddress = config.baseAddress;
/*   res->mapSize = config.mapSize; */
  res.trigFrequency = config.trigFrequency;
  res.resetFrequency = config.resetFrequency;
  return res;
}
